package controller;

import java.io.ByteArrayInputStream;
import java.io.IOException;
import java.io.InputStream;

import javax.sound.sampled.AudioFormat;
import javax.sound.sampled.AudioInputStream;
import javax.sound.sampled.TargetDataLine;

public class WaveCleaner extends AudioInputStream {

	boolean bigEndian;
	int bytes = 0;
	float sampleRate;
	short freqRing[];
	int freqRingPos = 0;

	// we calculate an approximate arc length of the sine wave
	// using the formula int [0,2pi] sqrt(cos(x)^2 + 1)
	// and see whether we can match it
	double arcLen = 7.640395578;
	//double arcLen = 6.8;
	
	// this determines the greatest arc that we won't match (cutoff*arcLen)
	// ideally this would be 16384 (if the samples were loud enough, ignores most noise)
	// or close to 0 (if there was no noise whatsoever)
	int amplitudeCutoff = 8000;
	float meterRingLambda = 1f; 
	
	short meterRing[];
	int meterRingPos = 0;
	int arcLenAbsolute = 0;
	
	// "fix" set bytes with 1-2 samples of "silence" in the middle by counting how
	// far we are in a machine bit - if between 0.3 and 1 of it, set bits to 1
	int outputBitSamples = 0;
	int samplesPerBit;

	AudioInputStream delegate;

	public WaveCleaner(InputStream stream, AudioFormat format, long length) {
		super(stream, format, length);
		setupData(format);
	}

	public WaveCleaner(TargetDataLine target) {
		super(target);
		setupData(target.getFormat());
	}

	public WaveCleaner(AudioInputStream delegate) {
		super(new ByteArrayInputStream(new byte[0]), delegate.getFormat(),
				delegate.getFrameLength());
		this.delegate = delegate;

		setupData(delegate.getFormat());
	}

	private void setupData(AudioFormat format) {
		if (format.getChannels() != 1 || format.getSampleSizeInBits() != 16) {
			throw new IllegalArgumentException("Require mono, signed, 16-bit");
		}

		System.out.println("big endian: "+format.isBigEndian());
		bigEndian = format.isBigEndian();
		bytes = format.getSampleSizeInBits();
		sampleRate = format.getSampleRate();

		// initialize frequency ring - frequency is 4800Hz
		// we generate two iterations
		freqRing = new short[(int) (sampleRate / 2400)];
		for (int i = 0; i < freqRing.length; i++) {
			freqRing[i] = (short) (Math.sin(4 * i * Math.PI / freqRing.length) * 32767); // 2*2*i
		}

		meterRing = new short[(int) (sampleRate*meterRingLambda / 4800)];
		
		// calculate the absolute arc length (for the meter ring)
		// also include the cutoff
		arcLenAbsolute = (int)(arcLen*amplitudeCutoff*(meterRing.length/(sampleRate/4800)));
		
		samplesPerBit = (int)(sampleRate/1200);
		
		System.out.println("frequency ring has " +freqRing.length+" samples, meter ring "+meterRing.length);
	}

	@Override
	public int read(byte[] b, int off, int len) throws IOException {
		byte[] samples = new byte[len];
		int read = (delegate != null) ? delegate.read(samples, 0, len) : super
				.read(samples, 0, len);

		for (int i = 0; i < read; i += 2) {
			short s = (short) (bigEndian ? ((samples[i] << 8) | ((samples[i + 1]+256)&0xff))
					: ((samples[i + 1] << 8) | ((samples[i]+256)&0xff)));

			// convert s according to heuristic
			meterRing[meterRingPos++] = s;			
			long arcLenMeter = 0;
			for (int j = 1; j < meterRing.length; j++) {
				arcLenMeter += Math.abs(meterRing[j-1]-meterRing[j]);
			}

			boolean set = arcLenMeter > arcLenAbsolute;

//			// if we're between 33% and 100% of a set bit and we generated a 0 sample, set it
//			if (outputBitSamples > samplesPerBit/3)
//				set = true;
//			
//			// increase or reset number of "recently set bits"
//			if (set)
//				outputBitSamples++;
//			else
//				outputBitSamples = 0;
//			
//			// roll over
//			if (outputBitSamples == samplesPerBit)
//				outputBitSamples = 0;
//			
//			
			s = set ? freqRing[freqRingPos++] : 0;

			if (freqRingPos == freqRing.length || !set)
				freqRingPos = 0;

			if (meterRingPos == meterRing.length)
				meterRingPos = 0;
		
			// save back - directly to b
			b[i + off] = (byte) (bigEndian ? (s >> 8) : (s & 0xff));
			b[i + 1 + off] = (byte) (bigEndian ? (s & 0xff) : (s >> 8));
		}

		return read;
	}

	// TODO: same for other methods (available, ...)

	@Override
	public void close() throws IOException {
		if (delegate != null)
			delegate.close();
		else
			super.close();
	}

	@Override
	public String toString() {
		return super.toString() + "WAVE Cleaner v2.5 - now with a mathematical background";
	}
/* old stuff
 *  //(short) (s < 0 ? 1 - s : s);

			long avg = 0;
			for (int j = 0; j < meterRing.length; j++) {
				avg += meterRing[j];
			}
			avg /= meterRing.length;

			long var = 0;
			for (int j = 0; j < meterRing.length; j++) {
				var += (avg - meterRing[j]) * (avg - meterRing[j]);
			}
			var /= meterRing.length;
			
			long weightedDelta = 0;
			for (int j = 0; j < meterRing.length; j++) {
				short prev = meterRing[(j + meterRing.length - 1)%meterRing.length];
				short cur = meterRing[j];
				
				if (Math.abs(prev) > noiseCertain && Math.abs(cur) > noiseCertain)
					weightedDelta += 32768;
				else
					weightedDelta += Math.abs(cur-prev);
			}
			weightedDelta /= meterRing.length;
			
			
			boolean set = (avg > noiseCertain
					|| var > (silenceDeviationCertain * silenceDeviationCertain) 
					|| (avg > noiseAverage && var > (silenceDeviationAverage * silenceDeviationAverage)));
			
 */
}
