package controller;

import java.io.IOException;
import java.io.OutputStream;
import java.util.Arrays;

import model.Cassette;
import model.PData;

public class WaveEncoder {

	private Cassette cassette;
	private static byte[] silenceBit;
	private static byte[] signalBitLittleEndian;
	private static byte[] signalBitBigEndian;
	private byte[] signalBit;
	private boolean stopped = false;
	
	public synchronized void stop() {
		stopped = true;
	}
	
	static {
		silenceBit = new byte[80];
		Arrays.fill(silenceBit, (byte)0);
		
		signalBitLittleEndian = new byte[80];
		
		// generated sine wave with a frequency of 4800hz
		// samples are: -1, -0.7, 0, 0.7, 1, 1, 0.7, 0, -0.7, -1
		// -32768, -23167, 0, 23167, 32767 and back
		// 10000000 00000000 | <90> <127> 
		for (int i = 0; i < 4; i++) {
			signalBitLittleEndian[i*20+0] = 0;
			signalBitLittleEndian[i*20+1] = (byte)0x80;
			
			signalBitLittleEndian[i*20+2] = 0;
			signalBitLittleEndian[i*20+3] = (byte)0xa5;
			
			signalBitLittleEndian[i*20+4] = 0;
			signalBitLittleEndian[i*20+5] = 0;
			
			signalBitLittleEndian[i*20+6] = 0x7f;
			signalBitLittleEndian[i*20+7] = 0x5a;
			
			signalBitLittleEndian[i*20+8] = (byte)0xff;
			signalBitLittleEndian[i*20+9] = 0x7f;
			
			for (int j = 0; j < 5; j++) {
				signalBitLittleEndian[i*20+10+j*2] = signalBitLittleEndian[i*20+(4-j)*2];
				signalBitLittleEndian[i*20+10+j*2+1] = signalBitLittleEndian[i*20+(4-j)*2+1];
			}
		}
		
		signalBitBigEndian = new byte[80];
		
		for (int i = 0; i < 40; i++) {
			signalBitBigEndian[2*i] = signalBitLittleEndian[2*i+1];
			signalBitBigEndian[2*i+1] = signalBitLittleEndian[2*i];
		}
		
	}
	
	public WaveEncoder(Cassette cassette, boolean bigEndian) {
		this.cassette = cassette;
		this.signalBit = bigEndian?signalBitBigEndian:signalBitLittleEndian;
	}
	
	/*
	 * assert the output stream is 48000 hz, 2 bits per sample, signed, big endian
	 * 
	 * some info from rené
	 * 
	 * - the frequency used by the C7420 for the audio output is 4.8kHz
     * - silence means '0', tone means '1'
     * - baudrate: 1200
     * - 1 start bit, 8 data bits (lsb first), 2.5 stopbits
     * 
     * bitSamples = sampleRate/baudrate * bitsPerSample
	 */
	
	private void outputSignalBlock(OutputStream out) throws IOException  {
		// 15 000 samples (375 bit samples) of 1
		for (int i = 0; i < 375 && !stopped; i++)
			out.write(signalBit);
		
		// 20 samples (0.5 bit samples) of 0
		if (!stopped)
			out.write(silenceBit, 0, silenceBit.length/2);
	}
	
	private void outputSilence(OutputStream out) throws IOException  {
		// 24 000 samples (half a second) of 0
		for (int i = 0; i < 600 && !stopped; i++)
			out.write(silenceBit);
	}
	
	private void encodeData(OutputStream out, PData data) throws IOException {
		short[] deviceBytes = data.encodeFull();
		
		for (short s: deviceBytes) {
			if (stopped)
				break;
			
			s = (short)((s<<1)|1); // "life" bit
			for (int i = 0; i < 9; i++) {
				if ((s & 0x1) == 1) {
					// tone
					out.write(signalBit);
				}
				else {
					// silence
					out.write(silenceBit);
				}
				s >>= 1;
			}
			// longer silence
			out.write(silenceBit);
			out.write(silenceBit);
			out.write(silenceBit, 0, silenceBit.length/2);
		}
	}
	
	public void encodeSelectedTab(OutputStream out) throws Exception {
		outputSignalBlock(out);
		encodeData(out, cassette.getSelected());
		outputSilence(out);
		
	}
	
	public void encode(OutputStream out) throws Exception {

		for (PData d: cassette) {
			outputSignalBlock(out);
			encodeData(out, d);
			outputSilence(out);
			
			if (stopped)
				return;
		}
		
	}
	
}
