package controller;

import gui.CassetteLogger;

import java.text.DecimalFormat;
import java.text.NumberFormat;

import javax.sound.sampled.AudioFormat;
import javax.sound.sampled.AudioInputStream;
import javax.sound.sampled.AudioSystem;

import model.Cassette;
import model.PData;

public class WaveParser implements Demodulator {

	private static final int BLOCK_SIZE = 512;
	private static final int AVERAGE_SPAN = 256; // number of samples after which to send an update

	protected boolean stopped = false;
	protected Cassette cassette;
	
	
	public WaveParser(Cassette cassette) {
		this.cassette = cassette;
	}
	
	public synchronized void stop() {
		stopped = true;
	}
	
	public void parse(AudioInputStream sourceStream) throws Exception {
		CassetteLogger.info("Using demodulator: "+getName());
		//CassetteLogger.warn("This demodulator does not do pre-amplification!");
		int cutoff = cassette.getOptions().getSignalCutoff() * cassette.getOptions().getPreAmplification() / 32768;
		CassetteLogger.debug("Adjusted cutoff: "+cutoff);
		
		AudioFormat f = sourceStream.getFormat();
		
		// get the desired target (16-bit, mono, unsigned, big endian)
		AudioFormat targetFormat = new AudioFormat(f.getSampleRate(), 16, 1, true, f.isBigEndian());
		
		// convert if possible (throw if not)
		AudioInputStream i = AudioSystem.getAudioInputStream(targetFormat, sourceStream);
		
		// start reading - a block at a time, abort if we're told to
		byte[] block = new byte[BLOCK_SIZE];
		short[] convertedBlock = new short[BLOCK_SIZE/2+1];
		int numRead = 0;

		int c; // current data byte
		int databyte=0,mask=0; // bits are accumulated in databyte 
		int bitnum=0; // current bit we are working on
		float bitcnt=0F;
		int cnt=0; // offset in block
		int datacnt=0; // offset in wav data block  
		int numb0=0; // count number of zeros
		int numb1=0; // count number of ones
		int stat=0; //0=start, 1=leading zero's, 2=6kHz, 3=data, 4=trailing zero's
		int prevstat=0; // previous state
		int is0=1;
		int plength=0; // program length
		boolean newprog=true; //loading a new program
		
		float bytesamples = f.getSampleRate() / 1200F;
		short[] BINData;    //binary data from file
		int BINDataUsage=0; //space used in BINData Array
		BINData = new short[17000]; //16K RAM + some file overhead space
		byte[] nextbyte = new byte[2];		
		int low,high;		

		int[] meterSamples = new int[AVERAGE_SPAN];
		
		while (!stopped && (numRead = i.read(block)) > 0) {
			// transform to 16-bit shorts
			int x,n,p;
			for (x=0;x<numRead/2;x++) {
				nextbyte[0]=block[x*2];
				nextbyte[1]=block[x*2+1];
				if (f.isBigEndian()) {
					low = nextbyte[1] & 0xff;
					high = nextbyte[0] & 0xff;
				} else {
					low = nextbyte[0] & 0xff;
					high = nextbyte[1] & 0xff;
				} 
					
				n= ( high << 8 | low );				
				if ((n&0x8000)==0x8000) // negative number
				{
						p=-((~n)&0xFFFF)+1;
				}
					else
						p=n;
				convertedBlock[x]=(short)p;	
			}
			
			cnt=0;
			/* process one block */
			
			if (newprog==true) {
				CassetteLogger.info("Searching for program header...");
				newprog=false;
			}
					
			while(cnt < numRead/2)
			{
				c = convertedBlock[cnt];
				cnt++;
				datacnt++;
				
				// save meter data
				meterSamples[(datacnt-1) % AVERAGE_SPAN] = c<0?-c:c;
				
				if (datacnt % AVERAGE_SPAN == 0) {
					// calculate average and deviation
					long avg = 0;
					for (int val: meterSamples)
						avg += val;
					avg /= AVERAGE_SPAN;
					
					long deviation = 0;
					for (int val: meterSamples)
						deviation += (avg-val)*(avg-val);
					
					deviation /= AVERAGE_SPAN;
					deviation = (long)Math.sqrt(deviation);
					
					// update meters
					CassetteLogger.setMeter((int)(avg*100/32768));
					CassetteLogger.setAuxMeter((int)(deviation*100/65536));
				}
				
				
				if ((c > -cutoff) && (c < cutoff)) numb0++; else {numb1++; is0=1; numb0=0;}
				if (numb0 > 2) {is0=0; numb1=0;}
				if ((stat == 0) && (numb0 > 460)) {stat=1; numb1=0;}
				if ((stat <  2) && (numb1 > 400)) {stat=2; numb0=0;}
				if ((stat == 2) && (numb0 >   6)) {stat=3; numb0=0; numb1=0; bitnum=0; bitcnt=0;}
				if ((stat == 3) && (numb0 > 460)) stat=4;

				if (prevstat!=stat) {
					CassetteLogger.debug("State " + stat + " reached at position " + datacnt);
				}

				prevstat=stat;
				
				if (stat==3)
				{
					bitcnt++;
					if ((bitnum==0) && (is0==1) && (numb1==20)) {bitnum=1; bitcnt=0; databyte=0x00; mask=0x01;}
					if ((bitnum> 0) && (bitcnt>=bytesamples))
					{
						bitcnt=bitcnt-bytesamples;
						if ((bitnum < 9) && (is0==1)) databyte |= mask;
						mask <<= 1;
						if (bitnum== 9) BINData[BINDataUsage++]=(short)databyte;						
						if (bitnum==10) {bitnum=0; numb1=0;}
						else bitnum++;
					}
				}				

				/*
				- 256x 0xFF
				- 10x 0xD3
				- 0x20 (ascii value of 'SPACE') (type of CSAVE, for example for CSAVEM this would be the ascii value of 'M')
				- program name (6 bytes)
				- 9x 0x00
				- start address of Basic program in RAM (0xC1 0x88) (LSB first)
				- program length (2 bytes)
				 		
				 */
				
				if (BINDataUsage==256+10+1+6+9+2+2 && plength==0) {				
					String fname="";
					int t=0;
					while (t<=5 && BINData[t+256+10+1]!=0) {
						fname=fname+(char)BINData[t+256+10+1];
						t++;
					}				
	
					/*low = BINData[256+10+1+6+9+2] & 0xff;
					high = BINData[256+10+1+6+9+2+1] & 0xff;
					plength = ( high << 8 | low );
					
					CassetteLogger.info("Found program type: 0x" + Integer.toHexString(BINData[256+10]) + "\nFilename:" + fname + ", length:" + plength);*/
				}
				
				/*
				- 256x 0xFF
				- 10x 0xD3
				- 0x20 (ascii value of 'SPACE') (type of CSAVE, for example for CSAVEM this would be the ascii value of 'M')
				- program name (6 bytes)
				- 9x 0x00
				- start address of Basic program in RAM (0xC1 0x88) (LSB first)
				- program length (2 bytes)
				- checksum (2 bytes)
				- 128x 0xFF
				- 1x 0x00
				- BASIC lines
				- 10x 0x00		
				 */
				
				if (BINDataUsage>=256+10+1+6+9+2+2+2+128+1+plength+10) {			
					short[] binblock=new short[BINDataUsage]; 
					System.arraycopy(BINData,0,binblock,0, BINDataUsage);		
					CassetteLogger.debug("Prepared: " + BINDataUsage + " data bytes.");
					showHex(binblock,BINDataUsage);
					PData dataBlock = PData.fromDataBlock(binblock, cassette.getOptions().onlyImportRaw());
					if (dataBlock != null && 
							(!cassette.getOptions().ignoreUnrecognizedBlocks() || !"unknown".equals(dataBlock.getName()))) {
						cassette.addBlock(dataBlock, true);
						CassetteLogger.info("Finished reading "+dataBlock.getName()+"...");
					}
					else if (dataBlock != null) {
						CassetteLogger.info("Ignoring unrecognized block");
					}
					else {
						CassetteLogger.warn("Could not parse data!");
					}
					
					/* start a new saved file */
					newprog=true;
					databyte=0;
					mask=0; 
					bitnum=0;
					bitcnt=0F;
					cnt=0;
					datacnt=0;  
					numb0=0;
					numb1=0;
					stat=0;
					prevstat=0;
					is0=1;
					plength=0;				
					BINDataUsage=0;
				}			
			}
		}
		
		if (plength!=0) { // save program with missing bytes
			short[] binblock=new short[BINDataUsage]; 
			System.arraycopy(BINData,0,binblock,0, BINDataUsage);		
			CassetteLogger.debug("Prepared: " + BINDataUsage + " data bytes.");
			showHex(binblock,BINDataUsage);
			
			PData dataBlock = PData.fromDataBlock(binblock, cassette.getOptions().onlyImportRaw());
			if (dataBlock != null && 
					(!cassette.getOptions().ignoreUnrecognizedBlocks() || !"unknown".equals(dataBlock.getName()))) {
				cassette.addBlock(dataBlock, true);
				CassetteLogger.warn("Incomplete program "+dataBlock.getName() +" detected...");
			}
			else if (dataBlock != null) {
				CassetteLogger.info("Ignoring unrecognized block");
			}
			else {
				CassetteLogger.warn("Could not parse data!");
			}
		}
		
	}

	void showHex(short[] BINData,int BINDataUsage) {
		 int width=32;
		 int r;
		 int l=0;
		 int s=0;
		 String asc="";
		 String hex="";
		 String h="";
		 NumberFormat formatter = new DecimalFormat("00000");		 
		 
		for (r=0;r<BINDataUsage;r++) {
			h=Integer.toHexString(BINData[r]);
			if (h.length()==1) 
				hex=hex + "0" + h + " ";
			else
				hex=hex + h + " ";
			
			if (BINData[r]<32) 
				asc=asc+".";
			else
				asc=asc + (char)(BINData[r]);
			
			l++;
			
			if (l==width) {
				System.out.println(formatter.format(s) + ": " + hex + asc);
				s=s+l;
				l=0;
				hex="";
				asc="";
			}		
		}
		if (hex.length()!=0) {
			for (r=l;r<width;r++)
				hex=hex+"   ";
			System.out.println(formatter.format(s) + ": " + hex + asc);
		}	
		
	}
	
	@Override
	public String getName() {
		return "WaveParser (based on Rene's work)";
	}
	
	@Override
	public int getDefaultCutoff() {
		return 16384;
	}
	
	public void reset() {
		stopped = false;
	}
}
