package controller;

import gui.CassetteLogger;

import javax.sound.sampled.AudioFormat;
import javax.sound.sampled.AudioInputStream;
import javax.sound.sampled.AudioSystem;

import model.Cassette;
import model.PData;

public class WaveParser2 extends WaveParser {

	private static final int BLOCK_SIZE = 512;
	private static final int AVERAGE_SPAN = 256; // number of samples after which to send an update
	
	private static final double ARC_LEN = 7.640395578;
	
	private short[] meterSamples = new short[AVERAGE_SPAN];
	private int meterPos = 0;
	
	private short[] data;
	private int dataSize;
	
	protected void updateMeters(short val) {

		// save meter data
		meterSamples[meterPos % AVERAGE_SPAN] = (short)(val<0?-(val+1):val);
		meterPos++;
		
		if (meterPos % AVERAGE_SPAN == 0) {
			// calculate average and deviation
			long avg = 0;
			for (int s: meterSamples)
				avg += s;
			avg /= AVERAGE_SPAN;
			
			long deviation = 0;
			for (int s: meterSamples)
				deviation += (avg-s)*(avg-s);
			
			deviation /= AVERAGE_SPAN;
			deviation = (long)Math.sqrt(deviation);
			
			// update meters
			CassetteLogger.setMeter((int)(avg*100/32768));
			CassetteLogger.setAuxMeter((int)(deviation*100/65536));
		}
		
	}
	
	public WaveParser2(Cassette cassette) {
		super(cassette);
	}
	
	protected long averageAmplitude(short[] ring, int start, int num) {
		long result = 0;
		
		for (int n = start; n < start+num; n++) {
			result += ring[n];
		}
		result /= num;

		return result;
	}
	
	protected long deviation(short[] ring, int start, int num, long average) {
		double result = 0;
		for (int n = start; n < start+num; n++) {
			result += (average-ring[n])*(average-ring[n]);
		}
		
		return (long)Math.sqrt(result);
	}
	
	// TODO optimize by looking at replaced value in array
	protected int arcLength(short[] ring, int start, int num) {
		int result = 0;
		
		for (int n = 1; n < num; n++) {
			result += Math.abs(ring[(n+ring.length+start-1)%ring.length]-ring[(n+ring.length+start)%ring.length]);
		}
		
		return result;
	}
	
	protected int absoluteArcLength(int level, int splits) {
		return (int)(level*(4.0/splits)*ARC_LEN);
	}
	
	protected int averageLengths(int signal, int silence) {
		return (signal+2*silence)/3;
	}
	
	protected void initData() {
		data = new short[1024];
		dataSize = 0;
	}
	
	/**
	 * Add a byte to the data stream
	 * @param b
	 */
	protected void addByte(short b) {
		data[dataSize++] = b;
		if (dataSize == data.length) {
			short[] newData = new short[data.length*2];
			System.arraycopy(data, 0, newData, 0, data.length);
			data = newData;
		}
	}
	
	protected void processData() {
		short[] sizedData = new short[dataSize];
		System.arraycopy(data, 0, sizedData, 0, dataSize);
		
		PData modelData = PData.fromDataBlock(sizedData, cassette.getOptions().onlyImportRaw());
		
		if (modelData != null && (!cassette.getOptions().ignoreUnrecognizedBlocks() || !"unknown".equals(modelData.getName()))) {
			cassette.addBlock(modelData,true);
			CassetteLogger.info("Finished parsing "+modelData.getName()+"...");
		}
		else if (modelData != null)
			CassetteLogger.info("Ignored unrecognized block");
		else
			CassetteLogger.warn("Could not parse data!");
		
		initData();
	}
	
	public void parse(AudioInputStream sourceStream) throws Exception {
		CassetteLogger.info("Using demodulator: "+getName());
		
		meterSamples = new short[AVERAGE_SPAN];
		meterPos = 0;
		
		this.data = null;
		this.dataSize = 0;
		
		//PrintWriter log = new PrintWriter(new File("out.log"));
		
		AudioFormat f = sourceStream.getFormat();
		
		// get the desired target (16-bit, mono, unsigned, big endian)
		AudioFormat targetFormat = new AudioFormat(f.getSampleRate(), 16, 1, true, f.isBigEndian());
		
		// convert if possible (throw if not)
		AudioInputStream i = AudioSystem.getAudioInputStream(targetFormat, sourceStream);
		
		// start reading - a block at a time, abort if we're told to
		byte[] block = new byte[BLOCK_SIZE];
		short[] convertedBlock = new short[BLOCK_SIZE/2+1];
		int numRead = 0;

		int datapos=0; // number of samples read (position in the data stream)
		int databyte=0;  // bits are accumulated in databyte
		int mask=0; // bit mask
		int bitnum=0; // number of bits already read from a byte
		
		float bitSamples = f.getSampleRate() / 1200f; // samples in a bit
		float byteSamples = bitSamples*11.5f; // samples in a byte
		
		/*
		 * how this works:
		 * for each sample, we need to find out whether it marks silence or signal. for that purpose,
		 * we create a "backtracking" ring - its size is the number of samples in a bit (sampleRate/baudRate)
		 * split it by a constant (splits). the ring is actually a simple array with an index that wraps around
		 * so that we always overwrite the oldest bit when adding a new one.
		 * 
		 * in order to get more accurate signal and silence values (when calibrating), the actual size of the ring
		 * was increased to contain a whole bit (so it's sampleRate/baudRate).
		 * 
		 * for this ring, we calculate the arc length (the actual length of the sine wave, roughly 7.64*sum of absolute
		 * value deltas) and check whether it is higher than a certain threshold. the initial threshold considers everything
		 * larger than 1/4th amplitude (8192) on average over "littleRingSize" samples to be signal, everything below
		 * is considered silence.
		 * 
		 * the threshold is recalibrated two times using the averageLengths function,
		 * while reading the constant signal block and while reading the first 2.5 bits silence from a byte.
		 */
		
		int splits = 2; // portion of a bit which determines whether the current sample is noise or silence
		int littleRingSize = (int)(bitSamples/splits); // number of samples which influence this
		short[] bitSampleRing = new short[littleRingSize*splits]; // the array containing samples for the last bit
		int ringPos = 0;
		
		// 100% - 1000% amplification (so max 32768/11)
		int ampLevel = 32768*100/(100+cassette.getOptions().getPreAmplification());
		int signalArcLength = absoluteArcLength(ampLevel, splits);
		int silenceArcLength = absoluteArcLength(0, splits);
		
		int adjustedCutoff = cassette.getOptions().getSignalCutoff()*ampLevel/32768;
		
		int arcLenCutoff = absoluteArcLength(adjustedCutoff, splits);
		
		CassetteLogger.debug("Initial signal level: "+ampLevel+", cutoff: "+cassette.getOptions().getSignalCutoff()+" (adjusted: "+adjustedCutoff+")");
		
		int permittedErrorSamples = (int)(bitSamples/16);
		
		int state = 0;
		int prevstate = 0;
		int num0 = 0;
		int num1 = 0;
				
		//log.println("bit sample ring has size "+bitSampleRing.length);
		//log.println("one bit has "+bitSamples+" samples, one byte has "+byteSamples+" samples");
		
		initData();
		
		while (!stopped && (numRead = i.read(block)) > 0) {
			/* process one BLOCK_SIZE sized block (so we can abort at will, e.g. while recording) */
			// transform to 16-bit shorts
			short p = 0;

			for (int x = 0; x < numRead / 2; x++) {
				p = (short) (f.isBigEndian() ? ((block[x*2] << 8) | ((block[x*2+1]+0x100)&0xff))
						: ((block[x*2+1] << 8) | ((block[x*2]+0x100)&0xff)));
				
				convertedBlock[x] = p;
			
				updateMeters(convertedBlock[x]);
				
				bitSampleRing[ringPos++] = convertedBlock[x];
				datapos++;
				
				// calculate data for this piece
				int arcLen = arcLength(bitSampleRing, ringPos-littleRingSize, littleRingSize);

				boolean isSet = arcLen > arcLenCutoff;
				boolean end0 = false;
				boolean end1 = false;
				int lastNum = 0;
				
				if (isSet) {
					num1++;
					if (num1 > permittedErrorSamples) {
						lastNum = num0;
						end0 = lastNum>permittedErrorSamples;
						num0 = 0;
					}
						
				}
				else {
					num0++;
					if (num0 > permittedErrorSamples) {
						lastNum = num1;
						end1 = lastNum>permittedErrorSamples;
						num1 = 0;
					}
				}
				
				if (ringPos == bitSampleRing.length)
					ringPos = 0;

				// states:
				if (state == 0 && end0 && lastNum >= 4*byteSamples) { // initial state, 4x silence => state 1 
					state = 1;
				}
				else if (state < 2 && end1 && lastNum >= 8*byteSamples) { // initial state/1, 8x signal => state 2 (calibration block)
					// start new program
					
					state = 2;
					
					// we should now save "signal block" calibration data
					signalArcLength = arcLength(bitSampleRing, 0, bitSampleRing.length)/splits;
					arcLenCutoff = averageLengths(signalArcLength, silenceArcLength);
					//CassetteLogger.debug("Cutoff is now: "+arcLenCutoff);
				}
				else if (state == 2 && end0 && lastNum >= 2*bitSamples) { // state 2, at least 2x silence => state 3 (end of first calibration block, read bits)
					state=3;
					
					// we should now save "silence block" calibration data
					silenceArcLength = arcLength(bitSampleRing, 0, bitSampleRing.length)/splits;
					arcLenCutoff = averageLengths(signalArcLength, silenceArcLength);
					//CassetteLogger.debug("Cutoff is now: "+arcLenCutoff);
					//log.println("Calculated cutoff: "+arcLenCutoff+" (silence: "+silenceArcLength+", signal: "+signalArcLength+")");
					
					// re-init the byte counter
					databyte = 0;
					mask = 0x1;
					bitnum = 0;
					
					// and add the first 0xff byte
					addByte((short)0xff);
					
					//log.println("read first 0xff byte");
				}
				else if (state == 3 && (end0 || end1)) { // state 3, read remaining bits
					int numBits = Math.round(lastNum / bitSamples);
					
					//log.println(datapos+" got "+(lastNum / bitSamples)+" bits of "+(end1?"1":"0"));
					
					for (int bit = 0; bit < numBits; bit++) {
						if (bitnum == 0 && !end1) {
							// first bit isn't alive - we have a problem
							CassetteLogger.warn("Byte around position "+datapos+" starts incorrectly, assuming silence");
							break; // we assume these are the 2.5 bits of silence at the end - the next byte will start with an alive bit.
						}
						if (bitnum >= 10)
							continue; // discard empty bits
						
						if (end1)
							databyte |= mask;
						
						mask <<= 1;
						bitnum++;
						
						if (bitnum == 10) {
							// tenth bit, commit (the first bit is 1, the last bit is 0, best discard both)
							// assert 11th and 11.5th bit are 0 and captured by "end0"
							addByte((short)((databyte >> 1) & 0xff));
							//log.println(datapos+": "+((short)((databyte >> 1) & 0xff)));
							databyte = 0;
							mask = 0x1;
						}
					}
					
					if (bitnum >= 10) {
						if (bitnum >= 13) {
							CassetteLogger.warn("Byte ending at position "+datapos+" too long ("+(lastNum / bitSamples)+")");
						}
						
						bitnum = 0;
					}
				}
				else if (state >= 2 && (num0 > 2 * byteSamples || num1 > 2 * byteSamples)) { // silence or signal block!
					if (state == 3) { // add last 0 bit
						addByte((short)((databyte>>1) & 0xff));
					}
					
					if (dataSize > 0)
						processData();
					
					// new state
					state = 0;
				}

				if (state < 5 && state != prevstate) {
					CassetteLogger.debug("Entered state "+state+" at position "+datapos);
					//log.println("Entered state "+state+" at position "+datapos);
					prevstate = state;
				}
			}
		}
		
		if (state == 3) {
			// add 0 byte
			addByte((short)((databyte>>1) & 0xff));
		}
		
		if (dataSize > 0)
			processData();
		
		//log.close();
	}
	
	@Override
	public String getName() {
		return "WaveParser2 (Arc length analysis)";
	}
	
	@Override
	public int getDefaultCutoff() {
		return 16384;
	}

}
