package controller.loaders;

import gui.CassetteLogger;
import gui.widgets.CharacterComponent;

import java.awt.Color;
import java.awt.Image;
import java.awt.image.BufferedImage;
import java.awt.image.Raster;
import java.io.File;
import java.io.IOException;
import java.util.Arrays;

import javax.imageio.ImageIO;
import javax.swing.filechooser.FileFilter;
import javax.swing.filechooser.FileNameExtensionFilter;

import model.PImage;

import controller.FileAction;

public class ImageLoader extends FileAction {

	private static final int ROWS = 23;
	private static final int COLS = 40;
	
	private static final int CHAR_WIDTH = 8;
	private static final int CHAR_HEIGHT = 10;
	
	@Override
	protected FileFilter[] getFilters() {
		return new FileFilter[]{
				new FileNameExtensionFilter(
						"Supported image formats (*.png, *.jpg, *.gif, *.bmp)", 
						"png", "jpg", "gif", "bmp")
		};
	}

	@Override
	public String[] getMenuPath() {
		return new String[]{"Import", "Image"};
	}

	@Override
	public boolean isLoadingAction() {
		return true;
	}

	private GraphicsChar getBestMatchingChar(int[] data) {
		// assert data size is [CHAR_HEIGHT][CHAR_WIDTH]
		GraphicsChar g = new GraphicsChar();
		int[] charData = new int[CHAR_HEIGHT*CHAR_WIDTH];
		boolean[] charSet = new boolean[CHAR_HEIGHT*CHAR_WIDTH];
		
		for (int c = 1; c < 128; c++) {
			for (int y = 0; y < CHAR_HEIGHT; y++) {
				for (int x = 0; x < CHAR_WIDTH; x++) {
					charSet[y*CHAR_WIDTH+x] = 
						(CharacterComponent.graphicsMode[c][y] &
								(1<<(CHAR_WIDTH-1-x))) != 0; 
				}
			}
			
			for (int fg = 0; fg < 8; fg++) {
				for (int bg = 0; bg < 8; bg++) {
					if (fg == bg)
						continue; // this will be matched by a char anyway
					
					// get char data
					for (int i = 0; i < charSet.length; i++)
						charData[i] = machineColors[charSet[i]?fg:bg];
					
					// calculate the score - lowest score wins
					double score = 0;
					
					for (int i = 0; i < charSet.length; i++)
						score += Math.abs((charData[i]&0xff) - (data[i]&0xff)) + 
						Math.abs(((charData[i]>>8)&0xff) - ((data[i]>>8)&0xff)) + 
						Math.abs(((charData[i]>>16)&0xff) - ((data[i]>>16)&0xff));
					
					if (score < g.score) {
						//System.err.println("found improved char "+c+" with score "+score+" (fg: "+fg+", bg: "+bg+")");
						g.c = c;
						g.bg = bg;
						g.fg = fg;
						g.score = score;
					}
				}
			}
		}
		
		//System.err.println("char "+g.c+": score "+g.score);
		return g;
	}
	
	@Override
	public void start() {
		for (File f: source) {
			CassetteLogger.info("Importing "+f.getName()+"...");
			try {
				BufferedImage i = ImageIO.read(f);
				
				// resize if necessary
				Image si = i.getScaledInstance(COLS*CHAR_WIDTH, ROWS*CHAR_HEIGHT, Image.SCALE_SMOOTH);
				i = new BufferedImage(COLS*CHAR_WIDTH, ROWS*CHAR_HEIGHT, BufferedImage.TYPE_INT_RGB);
				i.createGraphics().drawImage(si, 0, 0, null);
				
				// get best matching characters for the given blocks
				GraphicsChar[] chars = new GraphicsChar[ROWS*COLS];
				int[] data = new int[CHAR_WIDTH*CHAR_HEIGHT];
								
				int[] tempData = new int[CHAR_WIDTH*CHAR_HEIGHT];
								
				//System.err.println("image has "+i.getData().getNumBands()+" bands");
				
				for (int y = 0; y < ROWS; y++) {
					for (int x = 0; x < COLS; x++) {
						Arrays.fill(data, 0);
						Raster r = i.getData();
						int bands = r.getNumBands();
						for (int band = 0; band < r.getNumBands(); band++) {
							tempData = r.getSamples(x*CHAR_WIDTH, y*CHAR_HEIGHT, CHAR_WIDTH, CHAR_HEIGHT, band, tempData);
							for (int j = 0; j < tempData.length; j++)
								data[j] |= (tempData[j]<<(8*(bands-1-band)));
						}
						
						//data = r.getSamples(x*CHAR_WIDTH, y*CHAR_HEIGHT, CHAR_WIDTH, CHAR_HEIGHT, 0, data);
						
						chars[y*COLS+x] = getBestMatchingChar(data);
					}
				}
				
				PImage im = new PImage();
				im.setName(f.getName().replaceFirst("\\..*$", ""));
				char[][] imChars = im.getChars();
				int[][] imColors = im.getColors();
				
				double totalScore = 0;
				for (int y = 0; y < ROWS; y++) {
					for (int x = 0; x < COLS; x++) {
						GraphicsChar c = chars[y*COLS+x];
						totalScore += c.score;
						imChars[y][x] = (char)c.c;
						imColors[y][x] = 0x88|(c.bg<<4)|c.fg; /* graphics mode / no blink / bg / fg */
					}
				}
						
				cassette.addBlock(im, true);
				CassetteLogger.info("Imported "+f.getName()+" (inaccuracy: "+totalScore+")");
			} catch (IOException e) {
				CassetteLogger.warn("Could not import "+f.getName()+": "+e.getMessage());
			}
		}
	}

	private class GraphicsChar {
		int c = 0;
		int fg = 0;
		int bg = 0;
		double score = Double.MAX_VALUE;
	}
	
	private static int[] machineColors = new int[] {
		Color.BLACK.getRGB(),
		Color.RED.getRGB(),
		Color.GREEN.getRGB(),
		Color.YELLOW.getRGB(),
		Color.BLUE.getRGB(),
		Color.MAGENTA.getRGB(),
		Color.CYAN.getRGB(),
		Color.WHITE.getRGB()
	};
}
