package gui;

import java.awt.Component;
import java.awt.Container;
import java.awt.Dimension;
import java.awt.Graphics2D;
import java.awt.LayoutManager2;
import java.awt.geom.Rectangle2D;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.LinkedList;

import javax.swing.JLabel;

public class EasyLayout implements LayoutManager2 {

	private int hgap;
	private int vgap;
	
	public EasyLayout(int hgap, int vgap) {
		this.hgap = hgap;
		this.vgap = vgap;
	}
	
	public EasyLayout() {
		this(5,5);
	}
	
	/**
	 * This component starts a new row.
	 */
	public static final String NEW_ROW = "n";
	
	/**
	 * This component fills up remaining horizontal space (on the row) when resizing.
	 */
	public static final String HORIZONTAL_FILL = "h";	

	/**
	 * This component fills up remaining vertical space when resizing.
	 */
	public static final String VERTICAL_FILL = "v";	
	
	/**
	 * The width of this component is ignored when calculating the preferred size
	 */
	public static final String IGNORE_HORIZONTAL_WIDTH = "i";
	
	// i hate swing layouts. that's why i wrote this one, which is a bit easier to use
	// it doesn't need the layout manager object, instead it employs a string constraint object
	// add components in the order they should appear, components will be laid out like this:
	// - the preferred size will be the sum of the preferred sizes of all components + gaps 
	// - elements are laid-out left to right, left aligned (but will be justified if there is at least one "fill"/row)
	// 
	
	private class ConstrainedComponent {
		Component c;
		boolean horizontalFill;
		boolean verticalFill;
		boolean ignoreHorizontalWidth;
	}
	
	
	private ArrayList<ArrayList<ConstrainedComponent>> components = new ArrayList<ArrayList<ConstrainedComponent>>();
	
	@Override
	public void addLayoutComponent(Component comp, Object constraints) {
		// System.err.println("adding constraint!");	
		
		if (constraints == null || !(constraints instanceof String))
			constraints = "";
		
		String s = (String)constraints;
		ArrayList<ConstrainedComponent> row;
		if (s.contains(NEW_ROW) || components.size() == 0) {
			components.add(new ArrayList<ConstrainedComponent>());
		}
		
		row = components.get(components.size()-1);
		
		ConstrainedComponent c = new ConstrainedComponent();
		c.c = comp;
		c.horizontalFill = s.contains(HORIZONTAL_FILL);
		c.verticalFill = s.contains(VERTICAL_FILL);
		c.ignoreHorizontalWidth = s.contains(IGNORE_HORIZONTAL_WIDTH);
		row.add(c);
		
		
	}

	@Override
	public float getLayoutAlignmentX(Container target) {
		return 0;
	}

	@Override
	public float getLayoutAlignmentY(Container target) {
		return 0;
	}

	@Override
	public void invalidateLayout(Container target) {
		// nothing to do here (yet)
	}

	@Override
	public Dimension maximumLayoutSize(Container target) {
		return preferredLayoutSize(target);
	}

	@Override
	public void addLayoutComponent(String name, Component comp) {
	// System.err.println("adding string!");	
	/* unused */}

	@Override
	public void layoutContainer(Container parent) {
		// System.err.println("layout");
		// horizontal alignment: done for each row separately, basically:
		// partition into one fixed area and more fill areas, subtract the width
		// of the fixed area from the total, split the remaining width equally among all
		// fill areas (proportionately to their preferred sizes)
		
		// the vertical layout works according to the same principle
		
		int width = parent.getWidth()-2*hgap;
		int height = parent.getHeight()-2*vgap;

		int vfillPreferred = 0;
		
		ArrayList<Integer> verticalSizes = new ArrayList<Integer>();
		
		HashMap<ArrayList<ConstrainedComponent>, Integer> verticalFill = new HashMap<ArrayList<ConstrainedComponent>, Integer>();
		int remainingHeight = height;
		
		// first pass - horizontal alignment and vertical fill detection
		for (ArrayList<ConstrainedComponent> row: components) {
			int rowHeight = 0;
			int hfillPreferred = 0;
			
			LinkedList<ConstrainedComponent> horizFill = new LinkedList<ConstrainedComponent>();
			int remainingWidth = width + hgap;
			for (ConstrainedComponent c: row) {
				// System.err.println("Component "+c.c.getClass()+": "+c.c.getPreferredSize()+" hack: "+preferredSizeHack(c.c));
				
				if (!c.horizontalFill) {
					remainingWidth -= preferredSizeHack(c.c).width;
				}
				else {
					horizFill.add(c);
					hfillPreferred += preferredSizeHack(c.c).width;
				}
				c.c.setSize(preferredSizeHack(c.c));
				
				if (c.c.getHeight() > rowHeight) {
					rowHeight = c.c.getHeight();
				}
				
				if (c.verticalFill)
					verticalFill.put(row, rowHeight);
				
				remainingWidth -= hgap;
			}
			
			if (verticalFill.containsKey(row)) {
				verticalFill.put(row, rowHeight);
				vfillPreferred += rowHeight;
			}
			else
				remainingHeight -= rowHeight + vgap;
			
			verticalSizes.add(rowHeight);
			
			for (ConstrainedComponent c: horizFill) {
				// FFS - why does a JLABEL HAVE A PREFERREDSIZE OF 0,0?!?!
				if (c.c.getWidth() == 0 && hfillPreferred == 0)
					c.c.setSize(remainingWidth, c.c.getHeight());
				else
					c.c.setSize(c.c.getWidth()*remainingWidth/hfillPreferred, c.c.getHeight());
			}
		}
		
		// verticalFill now contains all rows with a VERTICAL_FILL component, as well as the preferred row sizes
		// verticalSizes contains all the row sizes, with VERTICAL_FILL ignored
		
		// the next step is to replace the sizes in verticalSizes for rows where VERTICAL_FILL is set
		// by using verticalFill and vfillPreferred
			
		for (int i = 0; i < components.size(); i++) {
			ArrayList<ConstrainedComponent> row = components.get(i);
			
			if (verticalFill.containsKey(row))
				verticalSizes.set(i, verticalSizes.get(i)*remainingHeight / vfillPreferred);
		}
		
		// all sizes are set, now for the real layout
		// components within a row are aligned to the top, using baselines instead when possible
		int j = 0;
		int xpos = hgap;
		int ypos = vgap;
		for (ArrayList<ConstrainedComponent> row: components) {
			int rowSize = verticalSizes.get(j);
			xpos = hgap;
			for (ConstrainedComponent c: row) {
				int yoffset = baselineHack(c.c, c.c.getWidth(), rowSize);
				if (yoffset < 0)
					yoffset = rowSize;
				
				c.c.setLocation(xpos, ypos+rowSize-yoffset);
				if (c.verticalFill)
					c.c.setSize(c.c.getWidth(), rowSize);
				
				// System.err.println("Component "+c.c.getClass()+": "+xpos+"x"+ypos+", size: "+c.c.getWidth()+"x"+c.c.getHeight()+" yoffset: "+yoffset);
				
				xpos += hgap + c.c.getWidth();
			}
			
			ypos += rowSize + vgap;
			j++;
			// System.err.println("----");
		}
	}

	@Override
	public Dimension minimumLayoutSize(Container parent) {
		return preferredLayoutSize(parent);
	}

	@Override
	public Dimension preferredLayoutSize(Container parent) {
		int x = 0;
		int y = 0;
		for (ArrayList<ConstrainedComponent> row: components) {
			int rowX = 0;
			int rowY = 0;
			
			for (ConstrainedComponent c: row) {
				rowX += c.ignoreHorizontalWidth?0:preferredSizeHack(c.c).width;
				if (preferredSizeHack(c.c).height > rowY)
					rowY = preferredSizeHack(c.c).height;
			}
			
			rowX += hgap*(row.size()-1);
			if (rowX > x)
				x = rowX;

			y += rowY;
		}

		if (components.size() > 0)
			y += vgap*(components.size()-1);
		
		return new Dimension(x+2*hgap,y+2*vgap);
	}

	@Override
	public void removeLayoutComponent(Component comp) {
		for (int row = 0; row < components.size(); row++) {
			ArrayList<ConstrainedComponent> componentRow = components.get(row);
			for (int col = 0; col < componentRow.size(); col++) {
				if (componentRow.get(col).c == comp) {
					componentRow.remove(col);
					col--;
				}
			}
			if (componentRow.size() == 0) {
				components.remove(row);
				row--;
			}
		}
	}

	// ... swing sucks.
	private Dimension preferredSizeHack(Component c) {
		Dimension d = c.getPreferredSize();
		
		if (d.height != 0 && d.width != 0)
			return d;
		
		if (c instanceof JLabel) {
			JLabel l = (JLabel)c;
			Rectangle2D r = l.getFont().getStringBounds(l.getText(), ((Graphics2D)l.getGraphics()).getFontRenderContext());
			return new Dimension((int)r.getWidth(), (int)r.getHeight());
		}
		
		return d;
	}
	
	private int baselineHack(Component c, int width, int height) {
		if (c instanceof JLabel) {
			return c.getBaseline(width, height);
		}
		else
			return -1;
	}
	
}
