package gui;

import java.awt.Component;
import java.awt.Dimension;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.io.IOException;
import java.io.InputStream;
import java.util.HashMap;

import javax.imageio.ImageIO;
import javax.swing.JFrame;
import javax.swing.JMenu;
import javax.swing.JMenuBar;
import javax.swing.JMenuItem;
import javax.swing.JOptionPane;
import javax.swing.JSeparator;
import javax.swing.KeyStroke;

import main.Main;
import model.Cassette;
import model.PData;
import controller.Demodulator;
import controller.IOAction;
import controller.WaveParser;
import controller.WaveParser2;
import controller.loaders.AudioFileLoader;
import controller.loaders.CaptureLoader;
import controller.loaders.HexLoader;
import controller.loaders.ImageLoader;
import controller.loaders.NativeFileLoader;
import controller.loaders.TemplateLoader;
import controller.loaders.ZipFileLoader;
import controller.savers.AudioFileSaver;
import controller.savers.HexSaver;
import controller.savers.MigrationSaver;
import controller.savers.NativeFileSaver;
import controller.savers.PlaybackSaver;
import controller.savers.ZipFileSaver;

@SuppressWarnings("serial")
public class MainFrame extends JFrame implements ActionListener {

	private MainPanel p;
	private Cassette c;
	
	private HashMap<JMenuItem, IOAction> actions = new HashMap<JMenuItem, IOAction>();
	private HashMap<JMenuItem, ActionFilter> actionEnablers = new HashMap<JMenuItem, ActionFilter>();
	
	public MainFrame() {
		c = new Cassette();
		init();
	}
	
	private void init() {
		setTitle("Data manager for the Philips G7400 + C7420 BASIC module");

		try {
			InputStream i = Main.getResource("res/icon.png"); 
			if (i != null)
				setIconImage(ImageIO.read(i));
			
		} catch (IOException e) {
			// icon not available
		}
		
		p = new MainPanel(c);
		p.addActionListener(this);
		
		setContentPane(p);
		initMenu();
		
		setMinimumSize(new Dimension(780, 400));
		
		pack();
	}
	
	private void addAction(JMenu menu, IOAction action, char shortcut, ActionFilter f) {
		String[] path = action.getMenuPath();
		
		int i = 0;
		
		// add submenus in path
		while (i < path.length - 1) {
			boolean foundMenu = false;
			for (Component c: menu.getMenuComponents()) {
				if (c instanceof JMenu && ((JMenu)c).getText() == path[i]) {
					menu = (JMenu)c;
					foundMenu = true;
					break;
				}
			}
			if (!foundMenu) {
				JMenu subMenu = new JMenu(path[i]);
				subMenu.setMnemonic(path[i].charAt(0));
				menu.add(subMenu);
				menu = subMenu;
			}
			i++;
		}
		
		// add menu action
		JMenuItem menuItem = new JMenuItem(path[i]);
		if (shortcut != 0) {
			menuItem.setMnemonic(shortcut);
			menuItem.setAccelerator(KeyStroke.getKeyStroke("control "+shortcut));
		}
		else {
			menuItem.setMnemonic(path[i].charAt(0));
		}
		
		actions.put(menuItem, action);
		actionEnablers.put(menuItem, f);
		menuItem.addActionListener(this);

		// cassette and stop callback will be set by startAction
		
		menu.add(menuItem);
		
	}
	
	private void addAction(JMenu menu, IOAction action, char shortcut) {
		addAction(menu, action, shortcut, savedActionFilter);
	}
	
	private void initMenu() {
		JMenuBar b = new JMenuBar();
		
		// add the file menu and all import/export options
		JMenu fileMenu = new JMenu("File");
		fileMenu.setMnemonic('f');
		
		b.add(fileMenu);
		
		p.setMenu(b);
	
		addAction(fileMenu, new TemplateLoader(), 'N');
		addAction(fileMenu, new NativeFileLoader(), 'O');
		
		
		addAction(fileMenu, new NativeFileSaver(false), 'S');
		addAction(fileMenu, new NativeFileSaver(true), '\0');
		final MigrationSaver migrationSaver = new MigrationSaver(p  /* EWWW */);
		addAction(fileMenu, migrationSaver, 'M', new ActionFilter() {
			@Override
			boolean accept() {
				return migrationSaver.accept(c.getSelected());
			}
		});
		
		addAction(fileMenu, new AudioFileLoader(true, false), '\0');
		addAction(fileMenu, new CaptureLoader(false), 'T');
		addAction(fileMenu, new CaptureLoader(true), '\0');
		addAction(fileMenu, new ZipFileLoader(), '\0');
		addAction(fileMenu, new ImageLoader(), '\0');
		
		addAction(fileMenu, new AudioFileSaver(false), '\0');
		addAction(fileMenu, new AudioFileSaver(true), '\0');
		addAction(fileMenu, new PlaybackSaver(true), 'P'); // XXX only playback a single file? i don't know - it's easy to enable all anyway
		addAction(fileMenu, new ZipFileSaver(), '\0');
		
		addAction(fileMenu, new HexSaver(), '\0');
		addAction(fileMenu, new HexLoader(), '\0');
		
		fileMenu.add(new JSeparator());
		
		JMenuItem exitItem = new JMenuItem("Exit");
		exitItem.setMnemonic('x');
		exitItem.setAccelerator(KeyStroke.getKeyStroke("control Q"));
		exitItem.setActionCommand("exit");
		exitItem.addActionListener(this);
		
		fileMenu.add(exitItem);
		
		p.getOptions().setDemodulators(new Demodulator[]{new WaveParser2(c), new WaveParser(c)});
		
		setItemsEnabled();
		
		/* new
		 * open [native, zip, (audio file?)]
		 * import -> audio file
		 * import -> audio capture source
		 * save (current tab)
		 * save as
		 * export -> zip archive
		 * export -> audio file
		 * export -> audio file (selected tab only)
		 * export -> audio playback
		 * export -> audio playback (selected tab only)
		 * export -> migrated format (???)
		 * exit
		 */
	}
	
	protected void setItemsEnabled() {
		for (JMenuItem i: actions.keySet()) {
			if (!actions.get(i).isLoadingAction())
				i.setEnabled(actionEnablers.get(i).accept());
		}
	}
	
	@Override
	public void actionPerformed(ActionEvent e) {
		if (p.equals(e.getSource())) {
			// tabs changed, refresh menu!
			setItemsEnabled();
		}
		else if (actions.containsKey(e.getSource())) {
			p.startAction(actions.get(e.getSource()));	
		}
		else if ("exit".equals(e.getActionCommand())) {
			boolean anyModified = false;
			
			for (PData d: c) {
				anyModified |= d.isModified() || (d.getDiskFiles() == null);
			}
			
			if (!anyModified || JOptionPane.showConfirmDialog(this, "Some tabs still have unsaved changes, exit without saving?", 
						"Unsaved changes", JOptionPane.YES_NO_OPTION) == JOptionPane.YES_OPTION)
					System.exit(0);
		}
	}
	
	private abstract class ActionFilter {
		abstract boolean accept();
	}
	
	ActionFilter savedActionFilter = new ActionFilter() {
		@Override
		boolean accept() {
			return c.getSelected() != null;
		}
	};
}
