package gui;

import gui.widgets.ArrayEditor;
import gui.widgets.BasicEditor;
import gui.widgets.HexEditor;
import gui.widgets.ImageEditor;
import gui.widgets.ModificationListener;
import gui.widgets.StringEditor;
import gui.widgets.WidgetSyncListener;

import java.awt.BorderLayout;
import java.awt.Component;
import java.awt.Dimension;
import java.awt.FlowLayout;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.event.MouseEvent;
import java.awt.event.MouseListener;
import java.lang.reflect.InvocationTargetException;
import java.util.HashMap;
import java.util.LinkedList;
import java.util.concurrent.ConcurrentLinkedQueue;

import javax.swing.JButton;
import javax.swing.JLabel;
import javax.swing.JMenuBar;
import javax.swing.JOptionPane;
import javax.swing.JPanel;
import javax.swing.JScrollPane;
import javax.swing.JSplitPane;
import javax.swing.JTabbedPane;
import javax.swing.SwingUtilities;
import javax.swing.event.ChangeEvent;
import javax.swing.event.ChangeListener;

import model.Cassette;
import model.CassetteListener;
import model.PArray;
import model.PBasicProgram;
import model.PData;
import model.PImage;
import model.PMachineCode;
import model.PString;
import controller.IOAction;
import controller.OptionStore;
import controller.StopCallback;

public class MainPanel extends JPanel implements CassetteListener, WidgetSyncListener, ChangeListener, ActionListener, ModificationListener {
	
	private static final long serialVersionUID = 129012418626985897L;

	private HashMap<Class<? extends PData>, Class<? extends Component>> componentMap = 
		new HashMap<Class<? extends PData>, Class<? extends Component>>();
	
	private LinkedList<ActionListener> listeners = new LinkedList<ActionListener>();
	
	private JTabbedPane tabs;
	private Cassette cassette;
	private OptionPanel options;
	private IOAction lastAction;
	
	private ConcurrentLinkedQueue<IOAction> runningActions = new ConcurrentLinkedQueue<IOAction>();
	
	public MainPanel(Cassette cassette) {
		this.cassette = cassette;
		cassette.addListener(this);
		
		init();
	}
	
	/**
	 * Actions are used whenever the number of tabs changes.
	 */
	public void addActionListener(ActionListener l) {
		listeners.add(l);
	}
	
	public void removeActionListener(ActionListener l) {
		listeners.remove(l);
	}
	
	public void notifyListeners() {
		ActionEvent e = new ActionEvent(this, 0, "");
		for (ActionListener l: listeners)
			l.actionPerformed(e);
	}
	
	private void init() {
		setLayout(new BorderLayout());
		
		JSplitPane p = new JSplitPane();
		tabs = new JTabbedPane();
		tabs.setPreferredSize(new Dimension(500,300));
		
		options = new OptionPanel();
		cassette.setOptions(options);
		
		p.add(tabs, JSplitPane.RIGHT);
		p.add(options, JSplitPane.LEFT);
		
		add(p);
		
		tabs.addChangeListener(this);
		
		options.getStartLastAction().addActionListener(this);
		options.getCloseTabs().addActionListener(this);
		
		// add widget factories here
		
		componentMap.put(PArray.class, ArrayEditor.class);
		componentMap.put(PBasicProgram.class, BasicEditor.class);
		componentMap.put(PImage.class, ImageEditor.class);
		componentMap.put(PString.class, StringEditor.class);
		componentMap.put(PMachineCode.class, HexEditor.class);
	}
	
	private Component getWidget(Class<? extends PData> c) {
		Component result = null;
		
		if (componentMap.containsKey(c)) {
			try {
				result = componentMap.get(c).newInstance();
			} catch (InstantiationException e) {
				// do nothing
			} catch (IllegalAccessException e) {
				// do nothing
			}
		}
		
		if (result == null)
			result = new HexEditor();
		
		return result;
	}
	
	@Override
	public void sync() {
		// sync all tabs
		
		for (Component c: tabs.getComponents()) {
			if (c instanceof JScrollPane) {
				c = ((JScrollPane)c).getViewport().getView();
			}
			
			if (c instanceof PDataEditor)
				((PDataEditor)c).syncData();
		}
		
	}
	
	@Override
	public synchronized void blockAdded(PData data) {
		SwingUtilities.invokeLater(new Runnable() {
			private PData data;
			
			@Override
			public void run() {

				// block added by ioaction, add tab		
				Component c = getWidget(data.getClass());
				
				Component addMe = c;
				
				// only add scrollbars if not PDataEditor or requested
				if (!(c instanceof PDataEditor) || ((PDataEditor)c).useScrollbars())
					addMe = new JScrollPane(c);
				
				tabs.addTab(data.getName(), addMe);
				tabs.setTabComponentAt(tabs.getTabCount()-1, 
						new TabClosingPanel(addMe, data.getName(), data.getAddress(), data.getInitData()));
				
				// focus
				tabs.setSelectedIndex(tabs.getTabCount()-1);
				cassette.setSelected(tabs.getTabCount()-1);
				
				// add data to c
				if (c instanceof PDataEditor) {
					((PDataEditor)c).setModificationListener(MainPanel.this);
					((PDataEditor)c).setData(data);
				}
				
				notifyListeners();
			}
			
			public Runnable setData(PData data) {
				this.data = data;
				return this;
			}
		}.setData(data));

	}
	
	// methods for tab management
	public void removeTab(Component c) {
		int idx = tabs.indexOfComponent(c);
		cassette.remove(idx);
		tabs.remove(idx);
		
		notifyListeners();
		
		if (cassette.isEmpty() && lastAction != null && !lastAction.isLoadingAction())
			options.getStartLastAction().setEnabled(false);
	}

	public void moveTab(Component c) { // move a tab to the right
		int idx = tabs.indexOfComponent(c);
		int newIdx = (idx+1)%cassette.size();
		
		Component tabComponent = tabs.getTabComponentAt(idx);
		
		// move data
		PData data = cassette.remove(idx);
		cassette.add(newIdx, data);
		
		// move editor
		tabs.remove(c);
		tabs.insertTab(null, null, c, null, newIdx);
		tabs.setTabComponentAt(newIdx, tabComponent);
		
		// re-focus
		tabs.setSelectedIndex(newIdx);
	}
	
	public void renameTab(Component c, String newName) {
		int idx = tabs.indexOfComponent(c);
		cassette.get(idx).setName(newName);
	}
	
	public int setTabAddress(Component c, Integer address) {
		int idx = tabs.indexOfComponent(c);
		cassette.get(idx).setAddress(address);
		return cassette.get(idx).getAddress();
	}
	
	public void setTabInitData(Component c, String newInitData) {
		int idx = tabs.indexOfComponent(c);
		cassette.get(idx).setInitData(newInitData);
	}
	
	@Override
	public void stateChanged(ChangeEvent e) {
		if (tabs.getSelectedIndex() >= 0)
			cassette.setSelected(tabs.getSelectedIndex());
		
		notifyListeners();
	}
	
	public void setMenu(JMenuBar b) {
		add(b, BorderLayout.NORTH);
	}
	
	public OptionPanel getOptions() {
		return options;
	}
	
	
	// tab class
	private class TabClosingPanel extends JPanel implements ActionListener, MouseListener {
		private static final long serialVersionUID = -3118049219378876395L;
		
		private Component contents;
		private String title;
		private String address;
		private String initData;
		private JButton addressButton;
		private JButton nameButton;
		private JButton initDataButton;
		private JLabel changed;
		
		public TabClosingPanel(Component contents, String title, int address, String initData) {
			super(new FlowLayout(FlowLayout.LEFT, 0, 0));
			setOpaque(false);
		
			this.contents = contents;
			this.title = title;
			this.address = Integer.toHexString(address);
			this.initData = initData;
			while (this.address.length() < 4)
				this.address = "0"+this.address;
			
			init();
		}
		
		private void init() {
			changed = new JLabel("");
			add(changed);
			
			nameButton = new JButton(title);
			nameButton.setToolTipText("Double-click to rename");
			nameButton.setBorderPainted(false);
			nameButton.setContentAreaFilled(false);
			nameButton.setActionCommand("rename");
			nameButton.addMouseListener(this);
			
			add(nameButton);
			
			addressButton = new JButton(address);
			addressButton.setFont(
					addressButton.getFont().deriveFont(
							addressButton.getFont().getSize2D()*0.8f));
			addressButton.setToolTipText("Double-click to set start address");
			addressButton.setBorderPainted(false);
			addressButton.setContentAreaFilled(false);
			addressButton.setActionCommand("address");
			addressButton.addMouseListener(this);
			add(addressButton);
			
			initDataButton = new JButton("ⁱ"+initData);
			initDataButton.setFont(
					initDataButton.getFont().deriveFont(
							initDataButton.getFont().getSize2D()*0.8f));
			initDataButton.setToolTipText("Double-click to set init data");
			initDataButton.setBorderPainted(false);
			initDataButton.setContentAreaFilled(false);
			initDataButton.setActionCommand("initData");
			initDataButton.addMouseListener(this);
			add(initDataButton);
			
			JButton moveRight = new JButton("→");
			moveRight.setToolTipText("Move tab to the right");
			moveRight.setBorderPainted(false);
			moveRight.setContentAreaFilled(false);
			moveRight.setActionCommand("move");
			moveRight.addActionListener(this);
			add(moveRight);
			
			JButton b = new JButton("✖");
			b.setToolTipText("Close tab");
			b.setBorderPainted(false);
			b.setContentAreaFilled(false);
			b.setActionCommand("close");
			b.addActionListener(this);
			add(b);
		}
		
		@Override
		public void actionPerformed(ActionEvent e) {
			if ("move".equals(e.getActionCommand())) {
				MainPanel.this.moveTab(contents);
			} else { // close
				boolean isModified = cassette.get(MainPanel.this.tabs.indexOfComponent(contents)).isModified();
				boolean isUnnamed = cassette.get(MainPanel.this.tabs.indexOfComponent(contents)).getDiskFiles() == null;
				String reason = isUnnamed?"\nThis file has not yet beed saved to disk.":(isModified?"\nThis file has been modified since the last save.":"");
				
				if ((!isModified && !isUnnamed) ||
						JOptionPane.showConfirmDialog(this, "Close '"+title+"'?"+reason, "Close tab",
						JOptionPane.YES_NO_OPTION) == JOptionPane.YES_OPTION)
					MainPanel.this.removeTab(contents);	
			}
			
		}
		
		@Override
		public void mouseClicked(MouseEvent e) {
			if (e.getClickCount() == 2) { // double-click to make event occur
				if (nameButton.equals(e.getSource())) {
					String newName = JOptionPane.showInputDialog("Enter new name for "+title, title);
					if (newName != null) {
						title = newName;
						MainPanel.this.renameTab(contents, title);
						nameButton.setText(title);
					}
				}
				else if (addressButton.equals(e.getSource())) {
					String newAddress = JOptionPane.showInputDialog("Enter new start address for "+title, address);
					if (newAddress != null && newAddress.matches("[0-9a-fA-F]{1,4}")) {
						newAddress = Integer.toHexString(
								MainPanel.this.setTabAddress(contents, Integer.valueOf(newAddress, 16)));

						while (newAddress.length() != 4)
							newAddress = "0"+newAddress;
						
						address = newAddress;
						addressButton.setText(address);
					}
				}
				else if (initDataButton.equals(e.getSource())) {
					String newInitData = JOptionPane.showInputDialog("Enter new init data for "+title, initData);
					if (newInitData != null) {
						initData = newInitData;

						MainPanel.this.setTabInitData(contents, initData);
						
						initDataButton.setText("ⁱ"+initData);
					}
				}
			}
			else {
				// FOCUS TAB, OMG
				tabs.setSelectedIndex(tabs.indexOfComponent(contents));
			}
		}
		@Override
		public void mouseEntered(MouseEvent e) {}
		@Override
		public void mouseExited(MouseEvent e) {}
		@Override
		public void mousePressed(MouseEvent e) {}
		@Override
		public void mouseReleased(MouseEvent e) {}
		
		public void setChanged(boolean changed) {
			if (changed) {
				this.changed.setText("*");
			}
			else
				this.changed.setText("");
		}
	}
	
	public OptionStore getOptionStore() {
		return options;
	}

	public void startAction(IOAction action) {
		// start an action in a new thread
		// this is just a helper function called from MainFrame
		
		// do not run the same action twice!
		if (runningActions.contains(action)) {
			JOptionPane.showMessageDialog(this, "Operation already running!");
			return;
		}
		
		action.setController(cassette, null);
		// sync editors
		sync();
		
		if (action.setSource()) {
			ActionThread t = new ActionThread(action);
			action.setController(cassette, t);
			
			// clear log if necessary
			if (options.autoClearLog())
				CassetteLogger.clear();
				
			runningActions.add(action);
			new Thread(t).start();
		}
		
	}
	
	private class ActionThread implements Runnable, StopCallback, ActionListener {
		private IOAction action;
		private JLabel actionLabel;
		private JButton stopButton; 

		ActionThread(IOAction action) {
			this.action = action;
		}
		
		@Override
		public void run() {
			// add items to action panel
			try {
				SwingUtilities.invokeAndWait(new Runnable() {
					@Override
					public void run() {
						// TODO Auto-generated method stub


						JPanel p = options.getActionPanel();
						actionLabel = new JLabel(joinMenuPath(action.getMenuPath()));
						stopButton = new JButton("Stop");
						stopButton.addActionListener(ActionThread.this);

						if (!action.canStop())
							stopButton.setEnabled(false);

						p.add(actionLabel, EasyLayout.NEW_ROW+EasyLayout.HORIZONTAL_FILL);
						p.add(stopButton);
						options.validate();
					}
				});
			} catch (InterruptedException e) {
				// hope this won't happen
			} catch (InvocationTargetException e) {
				// hope this won't happen
			}
			
			action.start();
			
			if (!action.canStop()) { // stop not supported, we'll have to stop ourselves
				stop();
			}
		}
		
		@Override
		public void stop() {
			SwingUtilities.invokeLater(new Runnable() {
				public void run() {
				// remove from action panel (running actions)
				options.getActionPanel().remove(stopButton);			
				options.getActionPanel().remove(actionLabel);

				// update last running action
				options.setLastAction(actionLabel.getText());
				options.getStartLastAction().setEnabled(true);
				lastAction = action;

				// XXX this should be done by the ioaction - update "modified" status on tabs 
				modified();

				options.validate();
				runningActions.remove(action);
				}
			});
		}
		
		@Override
		public void actionPerformed(ActionEvent e) {
			stopButton.setEnabled(false); // no need to stop several times
			action.stop();
			stop();
		}
		
		private String joinMenuPath(String[] path) {
			if (path.length == 0)
				return "";

			StringBuffer b = new StringBuffer();
			
			b.append(path[0]);
			
			for (int i = 1; i < path.length; i++)
				b.append(" ▸ "+path[i]);
			
			return b.toString();
		}
	}
	
	@Override
	public void actionPerformed(ActionEvent e) {
		if (e.getSource().equals(options.getStartLastAction())) {
			startAction(lastAction);
		}
		else if (e.getSource().equals(options.getCloseTabs())) {
			if (JOptionPane.showConfirmDialog(this, "Really close all tabs?", "Close all tabs",
					JOptionPane.YES_NO_OPTION) == JOptionPane.YES_OPTION) {
				tabs.removeAll();
				cassette.clear();
				
				notifyListeners();
				
				if (lastAction != null && !lastAction.isLoadingAction())
					options.getStartLastAction().setEnabled(false);
			}
		}
	}
	
	
	@Override
	public void modified() {
		for (int i = 0; i < cassette.size(); i++)
			((TabClosingPanel)tabs.getTabComponentAt(i)).setChanged(cassette.get(i).isModified());
	}
	
	// XXX for migrator
	public PDataEditor getEditorTab(int index) {
		Component c = tabs.getComponentAt(index);
		
		if (c instanceof JScrollPane)
			c = ((JScrollPane)c).getViewport().getView();
		
		if (c instanceof PDataEditor)
			return (PDataEditor)c;
		
		return null;
	}
}
