package gui.editor;

import java.awt.Component;
import java.text.DecimalFormat;
import java.text.ParseException;
import java.util.ArrayList;
import java.util.Collections;

import javax.swing.DefaultCellEditor;
import javax.swing.JFormattedTextField;
import javax.swing.JTable;
import javax.swing.event.TableModelListener;
import javax.swing.table.DefaultTableCellRenderer;
import javax.swing.table.DefaultTableModel;

import model.PArray;

public abstract class ArrayTable extends JTable {
	private static final long serialVersionUID = -6460093934806630553L;

	public abstract String getDescription();
	
	protected TableModelListener listener;
	
	// return null if you don't want anything to be added
	protected abstract Long valueToLong(int row, int col);
	
	// set the values in the model
	protected abstract void setValues(ArrayList<Long> arr);
	
	// get column identifiers
	protected abstract Object[] getColumnIdentifiers();
	
	// get row identifier - override this in the 2d view
	protected Object getRowIdentifier(int row) {
		return getPosition(row, extraDimensions);
	}
	
	protected ArrayList<Integer> extraDimensions = new ArrayList<Integer>();
	
	public static String getPosition(int row, ArrayList<Integer> extraDimensions) {
		// return a dimension-dependent representation of the row number
		ArrayList<Integer> pos = new ArrayList<Integer>();
		ArrayList<Integer> reverseDimensions = new ArrayList<Integer>(extraDimensions);
		Collections.reverse(reverseDimensions);
		
		for (int d: reverseDimensions) {
			pos.add(row % d);
			row /= d;
		}
		
		pos.add(row);
		Collections.reverse(pos);
		
		StringBuffer result = new StringBuffer();
		
		for (int p: pos) {
			result.append(p);
			result.append(" ");
		}
		
		// assert pos.size() > 0
		result.deleteCharAt(result.length()-1);
		
		return result.toString();
	}
	
	public ArrayList<Long> getArray() {
		ArrayList<Long> arr = new ArrayList<Long>();
		
		DefaultTableModel m = (DefaultTableModel)getModel();
		
		for (int row = 0; row < m.getRowCount(); row++) {
			for (int col = 0; col < m.getColumnCount(); col++) {
				Long l = valueToLong(row, col);
				
				if (l != null)
					arr.add(l);
			}
		}
		
		return arr;
	}
	
	public void setArray(ArrayList<Long> arr, ArrayList<Integer> extraDimensions) {
		this.extraDimensions = extraDimensions;
		
		DefaultTableModel m = new DefaultTableModel() {
			private static final long serialVersionUID = -9182477619517593426L;

			@Override
			public boolean isCellEditable(int row, int column) {
				return column > 0;
			}
		};
		
		//m.setColumnIdentifiers(getColumnIdentifiers());
		
		for (Object o: getColumnIdentifiers()) {
			m.addColumn(o);
		}
		
		if (listener != null)
			getModel().removeTableModelListener(listener);
		
		setModel(m);
		getColumnModel().getColumn(0).setWidth(32);
		
		if (listener != null)
			m.addTableModelListener(listener);
		
		setValues(arr); // !!! also set column values
	}
	
	
	public ArrayTable() {
		//setTableHeader(new JTableHeader());
	}
	
	// make this return true if you want useColumns events.
	public boolean usesColumns() {
		return false;
	}
		

	public void setModelListener(TableModelListener listener) {
		if (this.listener != null && getModel() != null)
			getModel().removeTableModelListener(this.listener);
		
		this.listener = listener;
		
		if (this.listener != null)
			getModel().addTableModelListener(this.listener);
	}
	
	public abstract Object[] defaultRow();
	
	public void renumber() {
		DefaultTableModel m = (DefaultTableModel)getModel();
		
		for (int i = 0; i < m.getRowCount(); i++)
			m.setValueAt(getRowIdentifier(i), i, 0);
	}
	
	public void insertRow(boolean before) {
		int row = getSelectedRow()+(before?0:1);
		if (row == -1)
			row = 0;
		else if (!before && row == 0)
			row = getRowCount();
		
		((DefaultTableModel)getModel()).insertRow(row, defaultRow());
		
		renumber();
	}
	
	public void appendColumn() { // stupid²²²² java
		// default: do nothing
	}

	public void deleteColumns() {
		// default: do nothing
	}
	
	public void deleteRows() {
		DefaultTableModel model = (DefaultTableModel)getModel();
		
		int rowOffset = 0;
		for (int row: getSelectedRows()) {
			model.removeRow(row-rowOffset);
			rowOffset++;
		}
		
		renumber();
	}

	
	// actual array tables
	public static class NumberArrayTable extends ArrayTable {
		private static final long serialVersionUID = -3384392024733893154L;

		@Override
		public String getDescription() {
			return "Flat numeric array";
		}
		
		@Override
		protected Object[] getColumnIdentifiers() {
			return new Object[]{"Position", "Value"};
		}
		
		@Override
		protected void setValues(ArrayList<Long> arr) {
			DefaultTableModel m = (DefaultTableModel)getModel();
			getColumnModel().getColumn(1).setCellEditor(new NumberEditor(Double.class, 10, -1));
			getColumnModel().getColumn(1).setCellRenderer(new MildDoubleRenderer());
			
			int i = 0;
			for (long l: arr) {
				m.addRow(new Object[]{getRowIdentifier(i), PArray.toFloat(l)});
				i++;
			}
		}
		
		@Override
		protected Long valueToLong(int row, int col) {
			if (col != 1)
				return null;
			
			Object o = getModel().getValueAt(row, col);
			
			if (o == null || !(o instanceof Double))
				return PArray.fromFloat(0.0);
			else
				return PArray.fromFloat((Double)o);
		}
		
		@Override
		public Object[] defaultRow() {
			return new Object[]{"", 0.0}; 
		}
	}
	
	public static class Number2DArrayTable extends ArrayTable {
		private static final long serialVersionUID = 8609196897933416432L;


		@Override
		public String getDescription() {
			return "Matrix (2D numeric arrays only)";
		}
		
		@Override
		protected Object[] getColumnIdentifiers() {
			int n = extraDimensions.get(0);
			Object[] o = new Object[n+1];
			
			o[0] = ""; // first cell
			for (int i = 0; i < n; i++)
				o[i+1] = Integer.toString(i);
			
			return o;
		}
		
		@Override
		protected void setValues(ArrayList<Long> arr) {
			DefaultTableModel m = (DefaultTableModel)getModel();
			setDefaultEditor(Object.class, new NumberEditor(Double.class, 10, -1));
			setDefaultRenderer(Object.class, new MildDoubleRenderer());
			
			int col = 0;
			int row = 0;
			ArrayList<Double> tmp = new ArrayList<Double>();
			
			for (long l: arr) {
				tmp.add(PArray.toFloat(l));
				col++;
				if (col == extraDimensions.get(0)) {
					Object o[] = new Object[col+1];
					o[0] = Integer.toString(row);
					for (int i = 0; i < col; i++)
						o[i+1] = tmp.get(i);
					
					m.addRow(o);
					tmp.clear();
					col = 0;
					row++;
				}
			}
			
			// add last row
			if (tmp.size() > 0) {
				Object o[] = new Object[col+1];
				o[0] = Integer.toString(row);
				for (int i = 0; i < col; i++)
					o[i+1] = i < tmp.size()?tmp.get(i):0.0;
				
				m.addRow(o);
			}
		}
		
		@Override
		protected Long valueToLong(int row, int col) {
			if (col == 0)
				return null;
			
			Object o = getModel().getValueAt(row, col);
			
			//System.err.println("getting value from "+row+" "+col+": "+o.getClass()+": "+o);
			
			if (o == null || !(o instanceof Double))
				return PArray.fromFloat(0.0);
			else
				return PArray.fromFloat((Double)o);
		}

		@Override
		public boolean usesColumns() {
			return true;
		}
		
		@Override
		public Object[] defaultRow() {
			Object[] objs = new Object[getModel().getColumnCount()];
			objs[0] = "";
			
			for (int i = 0; i < objs.length-1; i++)
				objs[i+1] = 0.0;
			
			return objs; 
		}
		
		@Override
		protected Object getRowIdentifier(int row) {
			return Integer.toString(row);
		}
		
		
		@Override
		public void appendColumn() {
			DefaultTableModel m = (DefaultTableModel)getModel();
			
			m.addColumn(m.getColumnCount()-1);
		}
		
	}

	public static class StringTable extends ArrayTable {
		private static final long serialVersionUID = 527038576086925756L;

		@Override
		public String getDescription() {
			return "Flat string array";
		}
		
		@Override
		protected Object[] getColumnIdentifiers() {
			return new Object[]{"Position", "Address (hex)", "Length"};
		}
		
		@Override
		protected void setValues(ArrayList<Long> arr) {
			DefaultTableModel m = (DefaultTableModel)getModel();
			getColumnModel().getColumn(1).setCellEditor(new NumberEditor(Long.class, 16, 0xffff));
			getColumnModel().getColumn(2).setCellEditor(new NumberEditor(Long.class, 10, 0xff));
			
			getColumnModel().getColumn(1).setCellRenderer(new HexRenderer());
			
			int i = 0;
			for (long l: arr) {
				m.addRow(new Object[]{getRowIdentifier(i), PArray.stringAddress(l), PArray.stringLength(l)});
				i++;
			}
		}
		
		@Override
		protected Long valueToLong(int row, int col) {
			DefaultTableModel m = (DefaultTableModel)getModel();
			
			if (col != 0)
				return null;
			
			return PArray.fromAddrLen((Long)m.getValueAt(row, 1), (Long)m.getValueAt(row, 2));
		}
		
		@Override
		public Object[] defaultRow() {
			return new Object[]{"", (long)0, (long)0};
		}
	}
	
	public static class RawTable extends ArrayTable {
		private static final long serialVersionUID = -5282326476272790471L;

		@Override
		public String getDescription() {
			return "Flat raw array";
		}
		
		@Override
		protected Object[] getColumnIdentifiers() {
			return new Object[]{"Position", "Data (hex)"};
		}
		
		@Override
		protected void setValues(ArrayList<Long> arr) {
			DefaultTableModel m = (DefaultTableModel)getModel();
			getColumnModel().getColumn(1).setCellEditor(new NumberEditor(Long.class, 16, 0xffffffff));
			getColumnModel().getColumn(1).setCellRenderer(new HexRenderer());
			
			int i = 0;
			for (long l: arr) {
				m.addRow(new Object[]{getRowIdentifier(i), l});
				i++;
			}
		}
		
		@Override
		protected Long valueToLong(int row, int col) {
			if (col != 1)
				return null;
			
			return (Long)getValueAt(row, col);
			
		}
		
		@Override
		public Object[] defaultRow() {
			return new Object[]{"", 0l};
		}
	}
	
	public static class NumberEditor extends DefaultCellEditor {
		private static final long serialVersionUID = 5488269334290188570L;
		
		private JFormattedTextField ft;
		
		// radix ignored for double!
		public NumberEditor(Class<?> c, int radix, long max) {
			
			
			super(radix==16?new JFormattedTextField(new RadixFormat(radix, max, c == Double.class)) {
				private static final long serialVersionUID = 1L;

				@Override
				public void setText(String t) {
					boolean hack17 = false;
					
					if (t.length() == 0)
						t = "0";
					if (t.startsWith("0x")) {
						t = t.substring(2);
						hack17 = true;
					}
					
					super.setText("0x"+Long.toHexString(Long.parseLong(t, hack17?16:10)));
				}
			}:new JFormattedTextField(new RadixFormat(radix, max, c == Double.class)));
			
			ft = (JFormattedTextField)getComponent(); // XXX one hack here is worse than the other .. java :///

		}
		
		@Override
		public Object getCellEditorValue() {
			try {
				ft.commitEdit();
				//System.out.println("value: "+ft.getValue()+" : "+ft.getValue().getClass());
				return ft.getValue();
			} catch (ParseException e) {
				return null;
			}
		}
		
		@Override
		public boolean stopCellEditing() {
			return ft.isEditValid() && super.stopCellEditing();
		}
	}
	
	public static class HexRenderer extends DefaultTableCellRenderer {
		private static final long serialVersionUID = 2626826516492917019L;

		@Override
		public Component getTableCellRendererComponent(JTable table,
				Object value, boolean isSelected, boolean hasFocus, int row,
				int column) {
			return super.getTableCellRendererComponent(table, "0x"+Long.toHexString((Long)value), 
					isSelected, hasFocus, row, column);
		}
	}
	
	public static class MildDoubleRenderer extends DefaultTableCellRenderer {
		private static final long serialVersionUID = -6631402688152678358L;

		@Override
		public Component getTableCellRendererComponent(JTable table,
				Object value, boolean isSelected, boolean hasFocus, int row,
				int column) {
			return super.getTableCellRendererComponent(table,
					(value != null && value instanceof Double)?new DecimalFormat("#.#######").format((Double)value):value, 
					isSelected, hasFocus, row, column);
		}		
	}
	
}
