package model;

import gui.CassetteLogger;

import java.io.BufferedReader;
import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.io.OutputStream;
import java.io.PrintWriter;
import java.util.ArrayList;

public class PArray extends PData {

	/**
	 * 
	 */
	private ArrayList<Long> arr = new ArrayList<Long>(); 
	
	/**
	 * An array has at least one dimension (the first to be specified in a DIM). The length of the array in this dimension
	 * is variable for our purposes.
	 * Further dimensions are fixed and specified here.
	 */
	private ArrayList<Integer> extraDimensions = new ArrayList<Integer>();
	
	/**
	 * The "native" file format saves (floating point) numbers as such and strings/raw as hex numbers.
	 */
	private boolean isNumber = false;
	
	@Override
	public void fromInputStream(InputStream i) throws IOException {
		arr.clear();
		
		String d = decodeVerboseHeader(i);
		BufferedReader b = new BufferedReader(new InputStreamReader(i));
		
		if (d == null)
			d = b.readLine();
		
		isNumber = false;
		if (d != null) {
			// read extra dimensions
			// format: <dim1> <dim2> ...
			if (d.startsWith("num "))
				isNumber = true;
			else if (d.startsWith("str "))
				isNumber = false;
			else
				throw new IOException("Missing num/str qualifier for array");
			
			d = d.substring(4);
			
			
			if (d.length() > 0)
				for (String s: d.split(" ")) {
					try {
						extraDimensions.add(Integer.parseInt(s));
					} catch (NumberFormatException e) {
						throw new IOException("Could not parse dimensions: "+e.getMessage());
					}
				}
			
		}

		String line;
		
		while ((line = b.readLine()) != null) {
			arr.add(isNumber?fromFloat(Double.parseDouble(line)):Long.parseLong(line, 16));
		}
	}
	
	@Override
	public void toOutputStream(OutputStream o) throws IOException {
		encodeVerboseHeader(o);
		
		PrintWriter p = new PrintWriter(o);
		
		// encode dimensions
		p.print(isNumber?"num":"str");

		for (int d: extraDimensions)
			p.print(" "+Integer.toString(d));
		
		p.println();
		
		for (Long l: arr) {
			p.println(isNumber?Double.toString(toFloat(l)):Long.toHexString(l));
		}
		p.flush();
	}
	
	public ArrayList<Long> getArray() {
		return arr;
	}
	
	public ArrayList<Integer> getExtraDimensions() {
		return extraDimensions;
	}
	
	public boolean isNumber() {
		return isNumber;
	}
	
	public void setNumber(boolean isNumber) {
		this.isNumber = isNumber;
	}
	
	@Override
	public short[] encode() {
		int totalLength = 1;
		
		for (int d: extraDimensions)
			totalLength *= d;
		
		int firstDim = (arr.size() + totalLength - 1)/totalLength;
		
		totalLength *= firstDim;
		
		short[] data = new short[1+(extraDimensions.size()+1)*2+totalLength*4];
		data[0] = (short)(extraDimensions.size()+1);
		
		data[1] = (short)(firstDim & 0xff);
		data[2] = (short)((firstDim>>8) & 0xff);
		
		int i = 3;
		
		for (int d: extraDimensions) {
			data[i] = (short)(d & 0xff);
			data[i+1] = (short)((d>>8) & 0xff);
			
			i+=2;
		}
		
		for (Long addr: arr) {

			data[i] = (short)(addr & 0xff);
			data[i+1] = (short)(addr>>8 & 0xff);
			data[i+2] = (short)(addr>>16 & 0xff);
			data[i+3] = (short)(addr>>24 & 0xff);
			
			i+=4;
		}
		
		for (int j = 0; j < totalLength - arr.size(); j++) {
			// fill remaining slots with zeroes
			data[i] = data[i+1] = data[i+2] = data[i+3] = 0;
			i+=4;
		}
			
		
		// zeroes are filled in automatically at the end
		return data;
	}
	
	@Override
	public void decode(short[] data) {
		arr.clear();
		
		// all is clear now!
		if (data.length < 3)
			return;
		
		// first is the number of dimensions
		
		// then come pairs of sizes
		int dim = data[0];
		
		if (dim < 1) {
			CassetteLogger.warn(getName()+": Array has 0 dimensions, leaving empty!");
			extraDimensions.clear();
			return;
		}
		
		ArrayList<Integer> lengths = new ArrayList<Integer>();
		
		long totalLength = 1;
		for (int i = 0; i < dim; i++) {
			lengths.add(data[2*i+1]|(data[2*i+2]<<8));
			totalLength *= data[2*i+1]|(data[2*i+2]<<8);
		}
		
		int offs = 1+lengths.size()*2;
		
		if (data.length-dim*2-1 != 4*totalLength) {
			CassetteLogger.warn(getName()+": wrong amount of elements, expected "+totalLength+", found "+((data.length-dim*2-1)/4.0));
			//long lengthRemainder = totalLength / lengths.get(0);
			totalLength = (data.length-dim*2-1)/4;
			// cap the size of the first dimension to whatever will fit in totalLength, leaving the other dimensions untouched 
			//lengths.set(0, (int)((totalLength+(lengthRemainder-1))/lengthRemainder));
		}
		
		for (int i = 0; i < totalLength; i++) {
			arr.add(data[offs+4*i] | data[offs+4*i+1]<<8 | data[offs+4*i+2]<<16 | (long)(data[offs+4*i+3])<<24);
		}
		
		lengths.remove(0);
		extraDimensions = lengths;
	}
	
	@Override
	public String getDefaultExtension() {
		return "arr";
	}
	
	@Override
	public char getFormat() {
		return 0xb7;
	}
	
	// fromDeviceFloat(short[] arr, int offset)
	public static double toFloat(long addr) {
		double result = 0;
		// first 23 bits are the mantissa
		result = (addr&0x7fffff);
		result /= 1<<23;
		result += 1;
		
		if (((addr>>23)&1) != 0) // 24th bit is the sign
			result = -result;
		
		// remaining values are the exponent - 0x81 is 0, lower values are negative exponents, higher are positive (i.e. biased)
		// exponent 0x0 = 0
		double exp = Math.pow(2, (addr>>24)-0x81);

		if (addr>>24 == 0)
			result = 0;
		else
			result *= exp;
		
		return result;
	}

	
	public static long fromFloat(double num) {
		boolean negative = num<0;
		
		if (negative)
			num = -num;
		
		int exp = 0;
		exp = Math.getExponent(num); // unbiased
		if (num == 0)
			exp = -0x81;
		else {
			num /= Math.pow(2, exp);
			num -= 1;
			num *= 1<<23;
		}
		
		int mantissa = (int)num;
		
		long ret = mantissa|(negative?0x1:0x0)<<23|((long)(exp+0x81)<<24);
		return ret;
	}
	
	public static long stringAddress(long l) {
		return l >> 16;
	}
	public static long stringLength(long l) {
		return l & 0xff;
	}
	
	public static long fromAddrLen(long addr, long len) {
		return (len & 0xff) | (0xff00) | ((addr << 16) & 0xffff0000);
	}
}
