package model;

import gui.CassetteLogger;

import java.io.File;
import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.io.PrintWriter;
import java.util.Arrays;
import java.util.regex.Matcher;
import java.util.regex.Pattern;


public abstract class PData {
	
	// diskFile is merely used as a hint, fromInputStream / toOutputStream do not access this file directly
	protected File[] diskFiles;
	protected boolean modified;
	
	protected String name;
	protected int address;
	protected String initData = "";
	
	public String getName() {
		return name;
	}

	public void setName(String name) {
		this.name = name;
	}

	public void setAddress(int address) {
		this.address = address;
	}
	
	public int getAddress() {
		return address;
	}
	
	public void setInitData(String initData) {
		this.initData = initData;
	}
	
	public String getInitData() {
		return initData;
	}
	
	public File[] getDiskFiles() {
		return diskFiles;
	}
	
	public void setDiskFiles(File[] diskFile) {
		this.diskFiles = diskFile;
	}
	
	public boolean isModified() {
		return modified;
	}
	
	public void setModified(boolean modified) {
		this.modified = modified;
	}
	
	
	//public static abstract ... fromInputStream(InputStream i);
	
	public static PData fromNative(String filename, InputStream i) throws IOException {
		String extension = filename.replaceFirst("^.*\\.([^.]+)$", "$1").toLowerCase();
		String file = filename.replaceFirst("^([^.]+).*", "$1");
		
		PData data = null;
		
		if ("txt".equals(extension) || "bas".equals(extension)) {
			data = new PBasicProgram();
		}
		else if ("arr".equals(extension)) {
			data = new PArray();
		}
		else if ("str".equals(extension)) {
			data = new PString();
		}
		else if ("img".equals(extension)) {
			data = new PImage();
		}
		
		// fallback
		if (data == null)
			data = new PRawData();
		
		data.setName(file);
		data.fromInputStream(i);
		
		i.close();
		
		return data;
	}
	
	public static PData fromDataBlock(short[] data) {
		return fromDataBlock(data, false);
	}
	
	public static PData fromDataBlock(short[] data, boolean returnRaw) {
		/* this function receives a header+data block (which could be e.g. one basic program, one string, whatever)
		 * from a WaveParser, looks at the header to determine what kind of a data block it is, parses out the name (if available),
		 * the length of the data block, its checksum, checks whether they are valid, then creates one of the model objects
		 * depending on the format, calls decode on the object and returns it.
		 */

		PData modelBlock = null;
		
		/*
		 * header format (for a BASIC program, ideally):
		 * - 256x 0xFF
		 * - 10x 0xD3
		 * - 0x20 (ascii value of 'SPACE') (type of CSAVE, for example for CSAVEM this would be the ascii value of 'M')
		 * - program name (6 bytes)
		 * - 1x 0x00
		 * - program init data (5 bytes) (the line the program will be started from, in ASCII)
		 * - 3x 0x00 (?)
		 * - start address of Basic program in RAM (0xC1 0x88) (LSB first)
		 * - program length (2 bytes)
		 * - checksum (2 bytes)
		 * - 128x 0xFF
		 * - 1x 0x00
		 * - BASIC lines
		 * - 10x 0x00		
		 */
		
		// ignore 0xff bytes - we allow for max "maxFaulty" consecutive faulty bytes.
		int maxFaulty = 16;
		int faultyBytes = 0;
		
		int type = 0;
		int length = 0;
		int checksum = 0;
		int startAddr = 0;
		StringBuffer nameBuffer = new StringBuffer();
		StringBuffer initDataBuffer = new StringBuffer();
		
		int pos = 0;
		
		while (faultyBytes < maxFaulty && pos < data.length) {
			if (data[pos] != 0xff)
				faultyBytes++;
			else
				faultyBytes = 0;
			
			pos++;
		}
		
		// skip at least 10-maxFaulty bytes, then read any more d3 bytes. this may not be ideal (since exactly those bytes may be)
		// damaged, but it's pretty hard to decide what to take otherwise
		int d3count = 0;

		if (faultyBytes > 9) {
			pos -= faultyBytes-9;
			faultyBytes = 9;
		}
		
		for (int i = pos-faultyBytes; i < pos; i++) {
			if (data[i] == 0xd3)
				d3count++;
		}
		
		int startPos = pos;
		while (pos < data.length && data[pos] == 0xd3) {
			d3count++;
			pos++;
		}
		
		//System.out.println("position: "+pos+" d3count: "+d3count);
		
		if (d3count > 0 && data.length-pos >= 22) {
			// process header
			type = data[pos++];
			for (int i = 0; i < 6 && data[pos+i] != 0; i++) {
				nameBuffer.append((char)data[pos+i]);
			}
			pos+=6;
			
			pos+=1; // 1x 0x00
			
			for (int i = 0; i < 5 && data[pos+i] != 0; i++) {
				initDataBuffer.append((char)data[pos+i]);
			}
			pos+=5;
			
			pos+=3; // 3x00
			
			startAddr = (data[pos+1] << 8) | data[pos];
			pos+=2;
			
			length = (data[pos+1] << 8) | data[pos];
			pos+=2;
			
			checksum = (data[pos+1] << 8) | data[pos];
			pos+=2;
			
			// read the next 0xff block (should be 128x, so we'll just skip it)
			if (data.length-pos >= 128)
				pos += 128;
		}
		else if (d3count == 0) {
			// revert to startPos
			pos = startPos;
			
			// skip until we get 6 0xffs in a row, then skip to the next 0
			int ffcount = 0;
			while (pos < data.length) {
				if (ffcount < 6) {
					if (data[pos] == 0xff)
						ffcount++;
					else
						ffcount = 0;
				}
				else {
					if (data[pos] == 0x0)
						break;
				}
				pos++;
			}
		}
		
		CassetteLogger.info("Found \""+(nameBuffer.length()==0?"<unknown>":nameBuffer.toString())+
				"\": type: 0x"+Integer.toHexString(type)+", address: 0x"+
				Integer.toHexString(startAddr)+", init data: "+initDataBuffer.toString()+", length: "+length);
		
		// skip the 0x0 byte
		pos++;
		
		// remove max 10 trailing zeroes
		int trailingZeroes = 0;
		for (int i = data.length-1; i >= 0 && data.length-i <= 10; i--)
			if (data[i] == 0)
				trailingZeroes++;
		
		// the stream is now set onto the 0x0 of the data block
		if (data.length - pos - trailingZeroes != length && length != 0) {
			CassetteLogger.warn(nameBuffer.toString()+": read "+(data.length-pos-trailingZeroes)+" bytes, expected "+length);
		}
		
		// check checksum
		int actualChecksum = 0;
		for (int i = pos; i < data.length; i++) {
			actualChecksum = (actualChecksum + data[i]) & 0xffff; 
		}
		
		if (actualChecksum != checksum) {
			CassetteLogger.warn(nameBuffer.toString()+": wrong checksum, expected 0x"+
					Integer.toHexString(checksum)+", got 0x"+Integer.toHexString(actualChecksum));
		}
		
		//System.out.println("type: "+Integer.toHexString(type));
		// create a different model object according to the type
		if (returnRaw)
			modelBlock = new PRawData();
		else {
			switch (type) {
			case 0x20:
				modelBlock = new PBasicProgram();
				break;
			case 0xb7:
				modelBlock = new PArray();
				break;
			case 0x53:
				modelBlock = new PImage(); 
				break;
			case 0x58:
				modelBlock = new PString(); 
				break;
			default:
				modelBlock = new PRawData();
			}
		}
		
		CassetteLogger.debug("Parsing as "+modelBlock.getClass().getSimpleName().substring(1)+"...");
		
		modelBlock.setName(nameBuffer.length() == 0?"unknown":nameBuffer.toString());
		modelBlock.setInitData(initDataBuffer.toString());
		modelBlock.setAddress(startAddr);
		
		short[] actualData;
		if (modelBlock instanceof PRawData) {
			actualData = data;
		}
		else {
			actualData = new short[data.length-pos-trailingZeroes];
			System.arraycopy(data, pos, actualData, 0, data.length-pos-trailingZeroes);
		}
		
		//System.out.println("trailing zeroes: "+trailingZeroes);
		
		modelBlock.decode(actualData);
		
		return modelBlock;
	}
	
	public short[] encodeFull() {
		short[] dataPart = encode();
		
		// add header to data part
		short[] data = new short[dataPart.length+256+10+1+6+9+6+128+1+10];
		

		Arrays.fill(data, 0, 256, (short)0xff); // 256 * 0xff
		Arrays.fill(data, 256, 266, (short)0xd3); // 10 * 0xd3
		data[266] = (short)getFormat(); // format
		for (int i = 0; i < 6; i++) {
			data[267+i] = (short)((name.length() > i) ? name.charAt(i) : (name.length() == i ? 0x0 : 0x20)); // name
		}
		
		data[273] = 0; // 1 * 0x00
		
		for (int i = 0; i < 5; i++) { // init data
			data[274+i] = (short)((initData.length() > i) ? initData.charAt(i) : 0x0); // name
		}
		
		Arrays.fill(data, 279, 282, (short)0); // 3 * 0x00
		
		data[282] = (short)(getAddress()&0xff);
		data[283] = (short)((getAddress()>>8)&0xff);
		
		int length = dataPart.length;
		
		data[284] = (short)(length&0xff);
		data[285] = (short)((length>>8)&0xff);
		
		int checksum = 0;
		for (short s: dataPart) {
			checksum = (checksum + s)&0xffff;
		}
		
		data[286] = (short)(checksum&0xff);
		data[287] = (short)((checksum>>8)&0xff);
		
		Arrays.fill(data, 288, 416, (short)0xff); // 128 * 0xff
		data[416] = 0; // 1 * 0x00
		
		System.arraycopy(dataPart, 0, data, 417, dataPart.length);
		
		// zeroes filled in automatically
		
		return data;
	}
	

	// default implementation - suggest override
	public void fromInputStream(InputStream i) throws IOException {
		String d = decodeVerboseHeader(i);
		
		byte[] buf = new byte[512];
		short[] data = new short[512];
		int offset = 0;
		int numRead = 0;
		
		if (d != null) {
			byte[] db = d.getBytes();
			
			for (int j = 0; j < db.length; j++) {
				data[j+offset] = (short)(((int)db[j]+256)%256);
			}
			offset += db.length;
		}
		
		while ((numRead = i.read(buf)) > 0) {
			for (int j = 0; j < numRead; j++) {
				data[j+offset] = (short)(((int)buf[j]+256)%256);
			}
			offset += numRead;
			
			if (data.length - offset < 512) {
				// increase data size
				short[] newData = new short[data.length*2];
				System.arraycopy(data, 0, newData, 0, data.length);
				data = newData;
			}
		}
		
		// get good data size
		short[] finalData = new short[offset];
		System.arraycopy(data, 0, finalData, 0, offset);
		data = finalData;
		
		decode(data);
	}
	
	// default implementation - suggest override
	public void toOutputStream(OutputStream o) throws IOException {
		encodeVerboseHeader(o);
		
		short[] data = encode(); 
		byte[] bytes = new byte[data.length];
		for (int i = 0; i < data.length; i++) {
			bytes[i] = (byte)data[i];
		}
		
		o.write(bytes);
	}
	
	// add a short header containing file name and address to the beginning of the data
	// in plain text, prefixed with an '
	// to be used when exporting to a native format (by toOutputStream)
	public void encodeVerboseHeader(OutputStream o) throws IOException {
		PrintWriter p = new PrintWriter(o);
		p.println("' address: 0x"+Integer.toHexString(getAddress())+" init: "+
				(getInitData().replace("\\", "\\\\").replace("\n", "\\n").replace(" ", "\\s"))+
				" name: "+(getName().replace("\\", "\\\\").replace("\n", "\\n")));
		p.flush();
	}
	
	// if the first string was NOT the header, return it
	// otherwise return null
	public String decodeVerboseHeader(InputStream i) throws IOException {
		
		StringBuffer b = new StringBuffer();
		
		int val;
		while ((val = i.read()) > 0) {
			if (val == '\n' || val == '\r') // XXX note that the returned string will contain a newline character but NOT MORE!!!
				break;
			b.append((char)val);
		}
		
		String l = b.toString();
		
		if (!l.matches("' address: 0x[0-9a-fA-F]{1,4} (?:init: [^ ]* )?name: .*"))
			return l+((val < 0)?"":(""+(char)val));
		else {
			Pattern p = Pattern.compile("' address: 0x([0-9a-fA-F]{1,4}) (?:init: ([^ ]*) )?name: (.*)");
			Matcher m = p.matcher(l);
			m.matches();
			
			setAddress(Integer.parseInt(m.group(1), 16));
			if (m.group(2) != null)
				setInitData(m.group(2).replace("\\s", " ").replace("\\n", "\n").replace("\\\\", "\\"));
			
			setName(m.group(3).replace("\\n", "\n").replace("\\\\", "\\"));
			
			return null;
		}
	}
	
	public abstract void decode(short[] data);
	public abstract short[] encode();
	
	public abstract String getDefaultExtension();
	
	public abstract char getFormat();
}
