package model;

import java.util.Arrays;

public class PImage extends PData {

	/*
	 * something about the graphics memory format:
	 * 
	 * first char of a row (column 0) has special significance:
	 * - 80 66 in text mode
	 * 
	 * 
	 * remaining columns: <ascii char> <formatting>
	 * formatting: 7 6 5 4 3 2 1 0 (lsb) [text mode]
	 *             | | | | | | | |
	 *             | | | | | +-+-+--- foreground color
	 *             | | | | +--------- !blink
	 *             | | | +----------- double height (start at row 1)
	 *             | | +------------- double width (start at column 1)
	 *             | +--------------- negative (bg on fg)
	 *             +----------------- 0 (controls graphics mode)
	 *             
	 * only chars 0-127 are "valid", other chars have special significance (control chars, see the manual)
	 *  
	 * formatting: 7 6 5 4 3 2 1 0 (lsb) [graphics mode]
	 *             | | | | | | | |
	 *             | | | | | +-+-+--- foreground color
	 *             | | | | +--------- !blink
	 *             | +-+-+----------- background color
	 *             +----------------- 1 (controls graphics mode)
	 * 
	 * chars 0x80 (and 0x7f?) changes the line's background color to the specified foreground (both hex digits are the same?)
	 */
	
	private final static int COLS = 40;
	private final static int ROWS = 23; 
	
	private char[][] chars = new char[ROWS][COLS];
	private int[][] colors = new int[ROWS][COLS];
	private int size = ROWS*COLS*2; // length of the data block (cols*rows*2)
	
	public PImage() {
		// init chars to all spaces
		for (int i = 0; i < ROWS; i++) {
			Arrays.fill(chars[i], ' ');
			chars[i][0] = (char)0x80; // bg
			
			Arrays.fill(colors[i], 0x08); // black
			colors[i][0] = 0x66; // cyan
		}
	}
	
	public char[][] getChars() {
		return chars;
	}
	
	public int[][] getColors() {
		return colors;
	}

	// raw file format
	
	@Override
	public short[] encode() {
		short[] data = new short[size];
		
		for (int i = 0; i < size/2; i++) {
			data[2*i] = (short)chars[i/COLS][i%COLS];
			data[2*i+1] = (short)colors[i/COLS][i%COLS];
		}
		
		return data;
	}
	
	@Override
	public void decode(short[] data) {
		// format: 80x23 bytes - whereas the first byte is the character, the second byte is the color
		// colors are (according to manual: #000 #f00 #0f0 #ff0 #00f #f0f #0ff #fff) - that's 3 bits for fg, 3 for bg,
		// ??? 1 for still/blinking - or 2 for bold/underline?
		int rows = (data.length/2+(COLS-1))/COLS;
		size = data.length;
		
		System.err.println("decoding "+rows+" rows...");
		
		chars = new char[rows][COLS];
		colors = new int[rows][COLS];
		
		for (int i = 0; i < data.length/2; i++) {
			chars[i/COLS][i%COLS] = (char)data[i*2];
			colors[i/COLS][i%COLS] = data[i*2+1];
		}
	}
	
	@Override
	public String getDefaultExtension() {
		return "img";
	}

	@Override
	public char getFormat() {
		return 'S';
	}
	
	@Override
	public int getAddress() {
		int a = super.getAddress();
		if (a == 0)
			return 0x8008;
		else
			return a;
	}
	
}
